package gnu.crypto.cipher;

// ----------------------------------------------------------------------------
// $Id: Serpent.java,v 1.2 2002/06/28 13:14:28 raif Exp $
//
// Copyright (C) 2002, Free Software Foundation, Inc.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 of the License or (at your option) any
// later version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
// more details.
//
// You should have received a copy of the GNU General Public License along with
// this program; see the file COPYING.  If not, write to the
//
//    Free Software Foundation Inc.,
//    59 Temple Place - Suite 330,
//    Boston, MA 02111-1307
//    USA
//
// As a special exception, if you link this library with other files to produce
// an executable, this library does not by itself cause the resulting
// executable to be covered by the GNU General Public License.  This exception
// does not however invalidate any other reasons why the executable file might
// be covered by the GNU General Public License.
// ----------------------------------------------------------------------------

import gnu.crypto.Registry;
import gnu.crypto.util.Util;

import java.security.InvalidKeyException;
import java.util.Collections;
import java.util.Iterator;

/**
 * <p>Serpent is a 32-round substitution-permutation network block cipher,
 * operating on 128-bit blocks and accepting keys of 128, 192, and 256 bits in
 * length. At each round the plaintext is XORed with a 128 bit portion of the
 * session key -- a 4224 bit key computed from the input key -- then one of
 * eight S-boxes are applied, and finally a simple linear transformation is
 * done. Decryption does the exact same thing in reverse order, and using the
 * eight inverses of the S-boxes.</p>
 *
 * <p>Serpent was designed by Ross Anderson, Eli Biham, and Lars Knudsen as a
 * proposed cipher for the Advanced Encryption Standard.</p>
 *
 * <p>Serpent can be sped up greatly by replacing S-box substitution with a
 * sequence of binary operations, and the optimal implementation depends
 * upon finding the fastest sequence of binary operations that reproduce this
 * substitution. This implementation uses the S-boxes discovered by
 * <a href="http://www.ii.uib.no/~osvik/">Dag Arne Osvik</a>, which are
 * optimized for the Pentium family of processors.</p>
 *
 * <p>References:</p>
 *
 * <ol>
 *    <li><a href="http://www.cl.cam.ac.uk/~rja14/serpent.html">Serpent: A
 *    Candidate Block Cipher for the Advanced Encryption Standard.</a></li>
 * </ol>
 *
 * @version $Revision: 1.2 $
 */
public class Serpent extends BaseCipher {

   // Constants and variables
   // -------------------------------------------------------------------------

   private int x0, x1, x2, x3, x4;

   private static final String NAME = "serpent";

   private static final int DEFAULT_KEY_SIZE = 16;
   private static final int DEFAULT_BLOCK_SIZE = 16;
   private static final int ROUNDS = 32;

   /** The fractional part of the golden ratio, (sqrt(5)+1)/2. */
   private static final int PHI = 0x9e3779b9;

   /**
    * KAT vector (from ecb_vk):
    * I=9
    * KEY=008000000000000000000000000000000000000000000000
    * CT=5587B5BCB9EE5A28BA2BACC418005240
    */
   private static final byte[] KAT_KEY =
         Util.toBytesFromString("008000000000000000000000000000000000000000000000");
   private static final byte[] KAT_CT =
         Util.toBytesFromString("5587B5BCB9EE5A28BA2BACC418005240");

   /** caches the result of the correctness test, once executed. */
   private static Boolean valid;

   // Constructor(s)
   // -------------------------------------------------------------------------

   /** Trivial zero-argument constructor. */
   public Serpent() {
      super(Registry.SERPENT_CIPHER, DEFAULT_BLOCK_SIZE, DEFAULT_KEY_SIZE);
   }

   // Class methods
   // -------------------------------------------------------------------------

   // Instance methods
   // -------------------------------------------------------------------------

   // java.lang.Cloneable interface implementation ----------------------------

   public Object clone() {
      return new Serpent();
   }

   // IBlockCipherSpi interface implementation --------------------------------

   public Iterator blockSizes() {
      return Collections.singleton(new Integer(DEFAULT_BLOCK_SIZE)).iterator();
   }

   public Iterator keySizes() {
      return new Iterator() {
         int i = 0;
         // Support 128, 192, and 256 bit keys.
         Integer[] keySizes = {
            new Integer(16), new Integer(24), new Integer(32)
         };

         public boolean hasNext() {
            return i < keySizes.length;
         }

         public Object next() {
            if (hasNext()) {
               return keySizes[i++];
            }
            return null;
         }

         public void remove() {
            throw new UnsupportedOperationException();
         }
      };
   }

   public Object makeKey(byte[] kb, int blockSize) throws InvalidKeyException {
      // Not strictly true, but here to conform with the AES proposal.
      // This restriction can be removed if deemed necessary.
      if (kb.length != 16 && kb.length != 24 && kb.length != 32) {
         throw new InvalidKeyException("Key length is not 16, 24, or 32 bytes");
      }
      SerpentKey key = new SerpentKey();

      // Here w is our "pre-key".
      int[] w = new int[4*(ROUNDS+1)];
      int i, j;
      for (i = 0, j = kb.length-4; i < 8 && j >= 0; i++) {
         w[i] = (kb[j  ] & 0xff) << 24 | (kb[j+1] & 0xff) << 16 |
                (kb[j+2] & 0xff) << 8  | (kb[j+3] & 0xff);
         j -= 4;
      }
      // Pad key if < 256 bits.
      if (i != 8) {
         w[i] = 1;
      }
      // Transform using w_i-8 ... w_i-1
      for (i = 8; i < 16; i++) {
         int t = w[i-8] ^ w[i-5] ^ w[i-3] ^ w[i-1] ^ PHI ^ (i-8);
         w[i] = t << 11 | t >>> 21;
      }
      // Translate by 8.
      for (i = 0; i < 8; i++) {
         w[i] = w[i+8];
      }
      // Transform the rest of the key.
      for (i = 8; i < w.length; i++) {
         int t = w[i-8] ^ w[i-5] ^ w[i-3] ^ w[i-1] ^ PHI ^ i;
         w[i] = t << 11 | t >>> 21;
      }

      // After these s-boxes the pre-key (w, above) will become the
      // session key (key).
      sbox3(w[  0], w[  1], w[  2], w[  3]);
      key.k0   = x0; key.k1   = x1; key.k2   = x2; key.k3   = x3;
      sbox2(w[  4], w[  5], w[  6], w[  7]);
      key.k4   = x0; key.k5   = x1; key.k6   = x2; key.k7   = x3;
      sbox1(w[  8],   w[  9],         w[ 10],        w[ 11]);
      key.k8   = x0; key.k9   = x1; key.k10  = x2; key.k11  = x3;
      sbox0(w[ 12],    w[ 13],        w[ 14],        w[ 15]);
      key.k12  = x0; key.k13  = x1; key.k14  = x2; key.k15  = x3;
      sbox7(w[ 16],    w[ 17],        w[ 18],        w[ 19]);
      key.k16  = x0; key.k17  = x1; key.k18  = x2; key.k19  = x3;
      sbox6(w[ 20],    w[ 21],        w[ 22],        w[ 23]);
      key.k20  = x0; key.k21  = x1; key.k22  = x2; key.k23  = x3;
      sbox5(w[ 24],    w[ 25],        w[ 26],        w[ 27]);
      key.k24  = x0; key.k25  = x1; key.k26  = x2; key.k27  = x3;
      sbox4(w[ 28],    w[ 29],        w[ 30],        w[ 31]);
      key.k28  = x0; key.k29  = x1; key.k30  = x2; key.k31  = x3;
      sbox3(w[ 32],    w[ 33],        w[ 34],        w[ 35]);
      key.k32  = x0; key.k33  = x1; key.k34  = x2; key.k35  = x3;
      sbox2(w[ 36],    w[ 37],        w[ 38],        w[ 39]);
      key.k36  = x0; key.k37  = x1; key.k38  = x2; key.k39  = x3;
      sbox1(w[ 40],    w[ 41],        w[ 42],        w[ 43]);
      key.k40  = x0; key.k41  = x1; key.k42  = x2; key.k43  = x3;
      sbox0(w[ 44],    w[ 45],        w[ 46],        w[ 47]);
      key.k44  = x0; key.k45  = x1; key.k46  = x2; key.k47  = x3;
      sbox7(w[ 48],    w[ 49],        w[ 50],        w[ 51]);
      key.k48  = x0; key.k49  = x1; key.k50  = x2; key.k51  = x3;
      sbox6(w[ 52],    w[ 53],        w[ 54],        w[ 55]);
      key.k52  = x0; key.k53  = x1; key.k54  = x2; key.k55  = x3;
      sbox5(w[ 56],    w[ 57],        w[ 58],        w[ 59]);
      key.k56  = x0; key.k57  = x1; key.k58  = x2; key.k59  = x3;
      sbox4(w[ 60],    w[ 61],        w[ 62],        w[ 63]);
      key.k60  = x0; key.k61  = x1; key.k62  = x2; key.k63  = x3;
      sbox3(w[ 64],    w[ 65],        w[ 66],        w[ 67]);
      key.k64  = x0; key.k65  = x1; key.k66  = x2; key.k67  = x3;
      sbox2(w[ 68],    w[ 69],        w[ 70],        w[ 71]);
      key.k68  = x0; key.k69  = x1; key.k70  = x2; key.k71  = x3;
      sbox1(w[ 72],    w[ 73],        w[ 74],        w[ 75]);
      key.k72  = x0; key.k73  = x1; key.k74  = x2; key.k75  = x3;
      sbox0(w[ 76],    w[ 77],        w[ 78],        w[ 79]);
      key.k76  = x0; key.k77  = x1; key.k78  = x2; key.k79  = x3;
      sbox7(w[ 80],    w[ 81],        w[ 82],        w[ 83]);
      key.k80  = x0; key.k81  = x1; key.k82  = x2; key.k83  = x3;
      sbox6(w[ 84],    w[ 85],        w[ 86],        w[ 87]);
      key.k84  = x0; key.k85  = x1; key.k86  = x2; key.k87  = x3;
      sbox5(w[ 88],    w[ 89],        w[ 90],        w[ 91]);
      key.k88  = x0; key.k89  = x1; key.k90  = x2; key.k91  = x3;
      sbox4(w[ 92],    w[ 93],        w[ 94],        w[ 95]);
      key.k92  = x0; key.k93  = x1; key.k94  = x2; key.k95  = x3;
      sbox3(w[ 96],    w[ 97],        w[ 98],        w[ 99]);
      key.k96  = x0; key.k97  = x1; key.k98  = x2; key.k99  = x3;
      sbox2(w[100],     w[101],        w[102],        w[103]);
      key.k100 = x0; key.k101 = x1; key.k102 = x2; key.k103 = x3;
      sbox1(w[104],     w[105],        w[106],        w[107]);
      key.k104 = x0; key.k105 = x1; key.k106 = x2; key.k107 = x3;
      sbox0(w[108],     w[109],        w[110],        w[111]);
      key.k108 = x0; key.k109 = x1; key.k110 = x2; key.k111 = x3;
      sbox7(w[112],     w[113],        w[114],        w[115]);
      key.k112 = x0; key.k113 = x1; key.k114 = x2; key.k115 = x3;
      sbox6(w[116],     w[117],        w[118],        w[119]);
      key.k116 = x0; key.k117 = x1; key.k118 = x2; key.k119 = x3;
      sbox5(w[120],     w[121],        w[122],        w[123]);
      key.k120 = x0; key.k121 = x1; key.k122 = x2; key.k123 = x3;
      sbox4(w[124],     w[125],        w[126],        w[127]);
      key.k124 = x0; key.k125 = x1; key.k126 = x2; key.k127 = x3;
      sbox3(w[128],     w[129],        w[130],       w[131]);
      key.k128 = x0; key.k129 = x1; key.k130 = x2; key.k131 = x3;

      return key;
   }

   public void
   encrypt(byte[] in, final int i, byte[] out, final int o, Object K, int bs) {
      SerpentKey key = (SerpentKey) K;
      //int x0, x1, x2, x3, x4;
      x3 = (in[i   ] & 0xff) << 24 | (in[i+ 1] & 0xff) << 16 |
           (in[i+ 2] & 0xff) <<  8 | (in[i+ 3] & 0xff);
      x2 = (in[i+ 4] & 0xff) << 24 | (in[i+ 5] & 0xff) << 16 |
           (in[i+ 6] & 0xff) <<  8 | (in[i+ 7] & 0xff);
      x1 = (in[i+ 8] & 0xff) << 24 | (in[i+ 9] & 0xff) << 16 |
           (in[i+10] & 0xff) <<  8 | (in[i+11] & 0xff);
      x0 = (in[i+12] & 0xff) << 24 | (in[i+13] & 0xff) << 16 |
           (in[i+14] & 0xff) <<  8 | (in[i+15] & 0xff);

      x0 ^= key.k0;
      x1 ^= key.k1;
      x2 ^= key.k2;
      x3 ^= key.k3;
      x3 ^= x0 ; x4 = x1 ;
      x1 &= x3 ; x4 ^= x2 ;
      x1 ^= x0 ; x0 |= x3 ;
      x0 ^= x4 ; x4 ^= x3 ;
      x3 ^= x2 ; x2 |= x1 ;
      x2 ^= x4 ; x4 ^= -1 ;
      x4 |= x1 ; x1 ^= x3 ;
      x1 ^= x4 ; x3 |= x0 ;
      x1 ^= x3 ; x4 ^= x3 ;
      x1 = (x1 << 13) | (x1 >>> 19);
      x4 ^= x1;
      x3 = x1 << 3;
      x2 = (x2 << 3) | (x2 >>> 29);
      x4 ^= x2;
      x0 ^= x2;
      x4 = (x4 << 1) | (x4 >>> 31);
      x0 ^= x3;
      x0 = (x0 << 7) | (x0 >>> 25);
      x3 = x4;
      x1 ^= x4;
      x3 <<= 7;
      x1 ^= x0;
      x2 ^= x0;
      x2 ^= x3;
      x1 = (x1 << 5) | (x1 >>> 27);
      x2 = (x2 << 22) | (x2 >>> 10);
      x1 ^= key.k4;
      x4 ^= key.k5;
      x2 ^= key.k6;
      x0 ^= key.k7;
      x4 =~ x4; x3 = x1;
      x1 ^= x4; x3 |= x4;
      x3 ^= x0; x0 &= x1;
      x2 ^= x3; x0 ^= x4;
      x0 |= x2; x1 ^= x3;
      x0 ^= x1; x4 &= x2;
      x1 |= x4; x4 ^= x3;
      x1 ^= x2; x3 |= x0;
      x1 ^= x3; x3 =~ x3;
      x4 ^= x0; x3 &= x2;
      x4 =~ x4; x3 ^= x1;
      x4 ^= x3;
      x0 = (x0 << 13) | (x0 >>> 19);
      x4 ^= x0;
      x3 = x0 << 3;
      x2 = (x2 << 3) | (x2 >>> 29);
      x4 ^= x2;
      x1 ^= x2;
      x4 = (x4 << 1) | (x4 >>> 31);
      x1 ^= x3;
      x1 = (x1 << 7) | (x1 >>> 25);
      x3 = x4;
      x0 ^= x4;
      x3 <<= 7;
      x0 ^= x1;
      x2 ^= x1;
      x2 ^= x3;
      x0 = (x0 << 5) | (x0 >>> 27);
      x2 = (x2 << 22) | (x2 >>> 10);
      x0 ^= key.k8;
      x4 ^= key.k9;
      x2 ^= key.k10;
      x1 ^= key.k11;
      x3 = x0 ;
      x0 = x0 & x2 ;
      x0 = x0 ^ x1 ;
      x2 = x2 ^ x4 ;
      x2 = x2 ^ x0 ;
      x1 = x1 | x3 ;
      x1 = x1 ^ x4 ;
      x3 = x3 ^ x2 ;
      x4 = x1 ;
      x1 = x1 | x3 ;
      x1 = x1 ^ x0 ;
      x0 = x0 & x4 ;
      x3 = x3 ^ x0 ;
      x4 = x4 ^ x1 ;
      x4 = x4 ^ x3 ;
      x3 = ~ x3 ;
      x2 = (x2 << 13) | (x2 >>> 19);
      x1 ^= x2;
      x0 = x2 << 3;
      x4 = (x4 << 3) | (x4 >>> 29);
      x1 ^= x4;
      x3 ^= x4;
      x1 = (x1 << 1) | (x1 >>> 31);
      x3 ^= x0;
      x3 = (x3 << 7) | (x3 >>> 25);
      x0 = x1;
      x2 ^= x1;
      x0 <<= 7;
      x2 ^= x3;
      x4 ^= x3;
      x4 ^= x0;
      x2 = (x2 << 5) | (x2 >>> 27);
      x4 = (x4 << 22) | (x4 >>> 10);
      x2 ^= key.k12;
      x1 ^= key.k13;
      x4 ^= key.k14;
      x3 ^= key.k15;
      x0 = x2 ;
      x2 = x2 | x3 ;
      x3 = x3 ^ x1 ;
      x1 = x1 & x0 ;
      x0 = x0 ^ x4 ;
      x4 = x4 ^ x3 ;
      x3 = x3 & x2 ;
      x0 = x0 | x1 ;
      x3 = x3 ^ x0 ;
      x2 = x2 ^ x1 ;
      x0 = x0 & x2 ;
      x1 = x1 ^ x3 ;
      x0 = x0 ^ x4 ;
      x1 = x1 | x2 ;
      x1 = x1 ^ x4 ;
      x2 = x2 ^ x3 ;
      x4 = x1 ;
      x1 = x1 | x3 ;
      x1 = x1 ^ x2 ;
      x1 = (x1 << 13) | (x1 >>> 19);
      x4 ^= x1;
      x2 = x1 << 3;
      x3 = (x3 << 3) | (x3 >>> 29);
      x4 ^= x3;
      x0 ^= x3;
      x4 = (x4 << 1) | (x4 >>> 31);
      x0 ^= x2;
      x0 = (x0 << 7) | (x0 >>> 25);
      x2 = x4;
      x1 ^= x4;
      x2 <<= 7;
      x1 ^= x0;
      x3 ^= x0;
      x3 ^= x2;
      x1 = (x1 << 5) | (x1 >>> 27);
      x3 = (x3 << 22) | (x3 >>> 10);
      x1 ^= key.k16;
      x4 ^= key.k17;
      x3 ^= key.k18;
      x0 ^= key.k19;
      x4 = x4 ^ x0 ;
      x0 = ~ x0 ;
      x3 = x3 ^ x0 ;
      x0 = x0 ^ x1 ;
      x2 = x4 ;
      x4 = x4 & x0 ;
      x4 = x4 ^ x3 ;
      x2 = x2 ^ x0 ;
      x1 = x1 ^ x2 ;
      x3 = x3 & x2 ;
      x3 = x3 ^ x1 ;
      x1 = x1 & x4 ;
      x0 = x0 ^ x1 ;
      x2 = x2 | x4 ;
      x2 = x2 ^ x1 ;
      x1 = x1 | x0 ;
      x1 = x1 ^ x3 ;
      x3 = x3 & x0 ;
      x1 = ~ x1 ;
      x2 = x2 ^ x3 ;
      x4 = (x4 << 13) | (x4 >>> 19);
      x2 ^= x4;
      x3 = x4 << 3;
      x1 = (x1 << 3) | (x1 >>> 29);
      x2 ^= x1;
      x0 ^= x1;
      x2 = (x2 << 1) | (x2 >>> 31);
      x0 ^= x3;
      x0 = (x0 << 7) | (x0 >>> 25);
      x3 = x2;
      x4 ^= x2;
      x3 <<= 7;
      x4 ^= x0;
      x1 ^= x0;
      x1 ^= x3;
      x4 = (x4 << 5) | (x4 >>> 27);
      x1 = (x1 << 22) | (x1 >>> 10);
      x4 ^= key.k20;
      x2 ^= key.k21;
      x1 ^= key.k22;
      x0 ^= key.k23;
      x4 = x4 ^ x2 ;
      x2 = x2 ^ x0 ;
      x0 = ~ x0 ;
      x3 = x2 ;
      x2 = x2 & x4 ;
      x1 = x1 ^ x0 ;
      x2 = x2 ^ x1 ;
      x1 = x1 | x3 ;
      x3 = x3 ^ x0 ;
      x0 = x0 & x2 ;
      x0 = x0 ^ x4 ;
      x3 = x3 ^ x2 ;
      x3 = x3 ^ x1 ;
      x1 = x1 ^ x4 ;
      x4 = x4 & x0 ;
      x1 = ~ x1 ;
      x4 = x4 ^ x3 ;
      x3 = x3 | x0 ;
      x1 = x1 ^ x3 ;
      x2 = (x2 << 13) | (x2 >>> 19);
      x0 ^= x2;
      x3 = x2 << 3;
      x4 = (x4 << 3) | (x4 >>> 29);
      x0 ^= x4;
      x1 ^= x4;
      x0 = (x0 << 1) | (x0 >>> 31);
      x1 ^= x3;
      x1 = (x1 << 7) | (x1 >>> 25);
      x3 = x0;
      x2 ^= x0;
      x3 <<= 7;
      x2 ^= x1;
      x4 ^= x1;
      x4 ^= x3;
      x2 = (x2 << 5) | (x2 >>> 27);
      x4 = (x4 << 22) | (x4 >>> 10);
      x2 ^= key.k24;
      x0 ^= key.k25;
      x4 ^= key.k26;
      x1 ^= key.k27;
      x4 = ~ x4 ;
      x3 = x1 ;
      x1 = x1 & x2 ;
      x2 = x2 ^ x3 ;
      x1 = x1 ^ x4 ;
      x4 = x4 | x3 ;
      x0 = x0 ^ x1 ;
      x4 = x4 ^ x2 ;
      x2 = x2 | x0 ;
      x4 = x4 ^ x0 ;
      x3 = x3 ^ x2 ;
      x2 = x2 | x1 ;
      x2 = x2 ^ x4 ;
      x3 = x3 ^ x1 ;
      x3 = x3 ^ x2 ;
      x1 = ~ x1 ;
      x4 = x4 & x3 ;
      x4 = x4 ^ x1 ;
      x2 = (x2 << 13) | (x2 >>> 19);
      x0 ^= x2;
      x1 = x2 << 3;
      x3 = (x3 << 3) | (x3 >>> 29);
      x0 ^= x3;
      x4 ^= x3;
      x0 = (x0 << 1) | (x0 >>> 31);
      x4 ^= x1;
      x4 = (x4 << 7) | (x4 >>> 25);
      x1 = x0;
      x2 ^= x0;
      x1 <<= 7;
      x2 ^= x4;
      x3 ^= x4;
      x3 ^= x1;
      x2 = (x2 << 5) | (x2 >>> 27);
      x3 = (x3 << 22) | (x3 >>> 10);
      x2 ^= key.k28;
      x0 ^= key.k29;
      x3 ^= key.k30;
      x4 ^= key.k31;
      x1 = x3 ;
      x3 = x3 & x0 ;
      x3 = x3 ^ x4 ;
      x4 = x4 & x0 ;
      x1 = x1 ^ x3 ;
      x3 = x3 ^ x0 ;
      x0 = x0 ^ x2 ;
      x2 = x2 | x1 ;
      x2 = x2 ^ x3 ;
      x4 = x4 ^ x0 ;
      x3 = x3 ^ x4 ;
      x4 = x4 & x2 ;
      x4 = x4 ^ x1 ;
      x1 = x1 ^ x3 ;
      x3 = x3 & x2 ;
      x1 = ~ x1 ;
      x3 = x3 ^ x1 ;
      x1 = x1 & x2 ;
      x0 = x0 ^ x4 ;
      x1 = x1 ^ x0 ;
      x3 = (x3 << 13) | (x3 >>> 19);
      x1 ^= x3;
      x0 = x3 << 3;
      x4 = (x4 << 3) | (x4 >>> 29);
      x1 ^= x4;
      x2 ^= x4;
      x1 = (x1 << 1) | (x1 >>> 31);
      x2 ^= x0;
      x2 = (x2 << 7) | (x2 >>> 25);
      x0 = x1;
      x3 ^= x1;
      x0 <<= 7;
      x3 ^= x2;
      x4 ^= x2;
      x4 ^= x0;
      x3 = (x3 << 5) | (x3 >>> 27);
      x4 = (x4 << 22) | (x4 >>> 10);
      x3 ^= key.k32;
      x1 ^= key.k33;
      x4 ^= key.k34;
      x2 ^= key.k35;
      x2 ^= x3 ; x0 = x1 ;
      x1 &= x2 ; x0 ^= x4 ;
      x1 ^= x3 ; x3 |= x2 ;
      x3 ^= x0 ; x0 ^= x2 ;
      x2 ^= x4 ; x4 |= x1 ;
      x4 ^= x0 ; x0 ^= -1 ;
      x0 |= x1 ; x1 ^= x2 ;
      x1 ^= x0 ; x2 |= x3 ;
      x1 ^= x2 ; x0 ^= x2 ;
      x1 = (x1 << 13) | (x1 >>> 19);
      x0 ^= x1;
      x2 = x1 << 3;
      x4 = (x4 << 3) | (x4 >>> 29);
      x0 ^= x4;
      x3 ^= x4;
      x0 = (x0 << 1) | (x0 >>> 31);
      x3 ^= x2;
      x3 = (x3 << 7) | (x3 >>> 25);
      x2 = x0;
      x1 ^= x0;
      x2 <<= 7;
      x1 ^= x3;
      x4 ^= x3;
      x4 ^= x2;
      x1 = (x1 << 5) | (x1 >>> 27);
      x4 = (x4 << 22) | (x4 >>> 10);
      x1 ^= key.k36;
      x0 ^= key.k37;
      x4 ^= key.k38;
      x3 ^= key.k39;
      x0 =~ x0; x2 = x1;
      x1 ^= x0; x2 |= x0;
      x2 ^= x3; x3 &= x1;
      x4 ^= x2; x3 ^= x0;
      x3 |= x4; x1 ^= x2;
      x3 ^= x1; x0 &= x4;
      x1 |= x0; x0 ^= x2;
      x1 ^= x4; x2 |= x3;
      x1 ^= x2; x2 =~ x2;
      x0 ^= x3; x2 &= x4;
      x0 =~ x0; x2 ^= x1;
      x0 ^= x2;
      x3 = (x3 << 13) | (x3 >>> 19);
      x0 ^= x3;
      x2 = x3 << 3;
      x4 = (x4 << 3) | (x4 >>> 29);
      x0 ^= x4;
      x1 ^= x4;
      x0 = (x0 << 1) | (x0 >>> 31);
      x1 ^= x2;
      x1 = (x1 << 7) | (x1 >>> 25);
      x2 = x0;
      x3 ^= x0;
      x2 <<= 7;
      x3 ^= x1;
      x4 ^= x1;
      x4 ^= x2;
      x3 = (x3 << 5) | (x3 >>> 27);
      x4 = (x4 << 22) | (x4 >>> 10);
      x3 ^= key.k40;
      x0 ^= key.k41;
      x4 ^= key.k42;
      x1 ^= key.k43;
      x2 = x3 ;
      x3 = x3 & x4 ;
      x3 = x3 ^ x1 ;
      x4 = x4 ^ x0 ;
      x4 = x4 ^ x3 ;
      x1 = x1 | x2 ;
      x1 = x1 ^ x0 ;
      x2 = x2 ^ x4 ;
      x0 = x1 ;
      x1 = x1 | x2 ;
      x1 = x1 ^ x3 ;
      x3 = x3 & x0 ;
      x2 = x2 ^ x3 ;
      x0 = x0 ^ x1 ;
      x0 = x0 ^ x2 ;
      x2 = ~ x2 ;
      x4 = (x4 << 13) | (x4 >>> 19);
      x1 ^= x4;
      x3 = x4 << 3;
      x0 = (x0 << 3) | (x0 >>> 29);
      x1 ^= x0;
      x2 ^= x0;
      x1 = (x1 << 1) | (x1 >>> 31);
      x2 ^= x3;
      x2 = (x2 << 7) | (x2 >>> 25);
      x3 = x1;
      x4 ^= x1;
      x3 <<= 7;
      x4 ^= x2;
      x0 ^= x2;
      x0 ^= x3;
      x4 = (x4 << 5) | (x4 >>> 27);
      x0 = (x0 << 22) | (x0 >>> 10);
      x4 ^= key.k44;
      x1 ^= key.k45;
      x0 ^= key.k46;
      x2 ^= key.k47;
      x3 = x4 ;
      x4 = x4 | x2 ;
      x2 = x2 ^ x1 ;
      x1 = x1 & x3 ;
      x3 = x3 ^ x0 ;
      x0 = x0 ^ x2 ;
      x2 = x2 & x4 ;
      x3 = x3 | x1 ;
      x2 = x2 ^ x3 ;
      x4 = x4 ^ x1 ;
      x3 = x3 & x4 ;
      x1 = x1 ^ x2 ;
      x3 = x3 ^ x0 ;
      x1 = x1 | x4 ;
      x1 = x1 ^ x0 ;
      x4 = x4 ^ x2 ;
      x0 = x1 ;
      x1 = x1 | x2 ;
      x1 = x1 ^ x4 ;
      x1 = (x1 << 13) | (x1 >>> 19);
      x0 ^= x1;
      x4 = x1 << 3;
      x2 = (x2 << 3) | (x2 >>> 29);
      x0 ^= x2;
      x3 ^= x2;
      x0 = (x0 << 1) | (x0 >>> 31);
      x3 ^= x4;
      x3 = (x3 << 7) | (x3 >>> 25);
      x4 = x0;
      x1 ^= x0;
      x4 <<= 7;
      x1 ^= x3;
      x2 ^= x3;
      x2 ^= x4;
      x1 = (x1 << 5) | (x1 >>> 27);
      x2 = (x2 << 22) | (x2 >>> 10);
      x1 ^= key.k48;
      x0 ^= key.k49;
      x2 ^= key.k50;
      x3 ^= key.k51;
      x0 = x0 ^ x3 ;
      x3 = ~ x3 ;
      x2 = x2 ^ x3 ;
      x3 = x3 ^ x1 ;
      x4 = x0 ;
      x0 = x0 & x3 ;
      x0 = x0 ^ x2 ;
      x4 = x4 ^ x3 ;
      x1 = x1 ^ x4 ;
      x2 = x2 & x4 ;
      x2 = x2 ^ x1 ;
      x1 = x1 & x0 ;
      x3 = x3 ^ x1 ;
      x4 = x4 | x0 ;
      x4 = x4 ^ x1 ;
      x1 = x1 | x3 ;
      x1 = x1 ^ x2 ;
      x2 = x2 & x3 ;
      x1 = ~ x1 ;
      x4 = x4 ^ x2 ;
      x0 = (x0 << 13) | (x0 >>> 19);
      x4 ^= x0;
      x2 = x0 << 3;
      x1 = (x1 << 3) | (x1 >>> 29);
      x4 ^= x1;
      x3 ^= x1;
      x4 = (x4 << 1) | (x4 >>> 31);
      x3 ^= x2;
      x3 = (x3 << 7) | (x3 >>> 25);
      x2 = x4;
      x0 ^= x4;
      x2 <<= 7;
      x0 ^= x3;
      x1 ^= x3;
      x1 ^= x2;
      x0 = (x0 << 5) | (x0 >>> 27);
      x1 = (x1 << 22) | (x1 >>> 10);
      x0 ^= key.k52;
      x4 ^= key.k53;
      x1 ^= key.k54;
      x3 ^= key.k55;
      x0 = x0 ^ x4 ;
      x4 = x4 ^ x3 ;
      x3 = ~ x3 ;
      x2 = x4 ;
      x4 = x4 & x0 ;
      x1 = x1 ^ x3 ;
      x4 = x4 ^ x1 ;
      x1 = x1 | x2 ;
      x2 = x2 ^ x3 ;
      x3 = x3 & x4 ;
      x3 = x3 ^ x0 ;
      x2 = x2 ^ x4 ;
      x2 = x2 ^ x1 ;
      x1 = x1 ^ x0 ;
      x0 = x0 & x3 ;
      x1 = ~ x1 ;
      x0 = x0 ^ x2 ;
      x2 = x2 | x3 ;
      x1 = x1 ^ x2 ;
      x4 = (x4 << 13) | (x4 >>> 19);
      x3 ^= x4;
      x2 = x4 << 3;
      x0 = (x0 << 3) | (x0 >>> 29);
      x3 ^= x0;
      x1 ^= x0;
      x3 = (x3 << 1) | (x3 >>> 31);
      x1 ^= x2;
      x1 = (x1 << 7) | (x1 >>> 25);
      x2 = x3;
      x4 ^= x3;
      x2 <<= 7;
      x4 ^= x1;
      x0 ^= x1;
      x0 ^= x2;
      x4 = (x4 << 5) | (x4 >>> 27);
      x0 = (x0 << 22) | (x0 >>> 10);
      x4 ^= key.k56;
      x3 ^= key.k57;
      x0 ^= key.k58;
      x1 ^= key.k59;
      x0 = ~ x0 ;
      x2 = x1 ;
      x1 = x1 & x4 ;
      x4 = x4 ^ x2 ;
      x1 = x1 ^ x0 ;
      x0 = x0 | x2 ;
      x3 = x3 ^ x1 ;
      x0 = x0 ^ x4 ;
      x4 = x4 | x3 ;
      x0 = x0 ^ x3 ;
      x2 = x2 ^ x4 ;
      x4 = x4 | x1 ;
      x4 = x4 ^ x0 ;
      x2 = x2 ^ x1 ;
      x2 = x2 ^ x4 ;
      x1 = ~ x1 ;
      x0 = x0 & x2 ;
      x0 = x0 ^ x1 ;
      x4 = (x4 << 13) | (x4 >>> 19);
      x3 ^= x4;
      x1 = x4 << 3;
      x2 = (x2 << 3) | (x2 >>> 29);
      x3 ^= x2;
      x0 ^= x2;
      x3 = (x3 << 1) | (x3 >>> 31);
      x0 ^= x1;
      x0 = (x0 << 7) | (x0 >>> 25);
      x1 = x3;
      x4 ^= x3;
      x1 <<= 7;
      x4 ^= x0;
      x2 ^= x0;
      x2 ^= x1;
      x4 = (x4 << 5) | (x4 >>> 27);
      x2 = (x2 << 22) | (x2 >>> 10);
      x4 ^= key.k60;
      x3 ^= key.k61;
      x2 ^= key.k62;
      x0 ^= key.k63;
      x1 = x2 ;
      x2 = x2 & x3 ;
      x2 = x2 ^ x0 ;
      x0 = x0 & x3 ;
      x1 = x1 ^ x2 ;
      x2 = x2 ^ x3 ;
      x3 = x3 ^ x4 ;
      x4 = x4 | x1 ;
      x4 = x4 ^ x2 ;
      x0 = x0 ^ x3 ;
      x2 = x2 ^ x0 ;
      x0 = x0 & x4 ;
      x0 = x0 ^ x1 ;
      x1 = x1 ^ x2 ;
      x2 = x2 & x4 ;
      x1 = ~ x1 ;
      x2 = x2 ^ x1 ;
      x1 = x1 & x4 ;
      x3 = x3 ^ x0 ;
      x1 = x1 ^ x3 ;
      x2 = (x2 << 13) | (x2 >>> 19);
      x1 ^= x2;
      x3 = x2 << 3;
      x0 = (x0 << 3) | (x0 >>> 29);
      x1 ^= x0;
      x4 ^= x0;
      x1 = (x1 << 1) | (x1 >>> 31);
      x4 ^= x3;
      x4 = (x4 << 7) | (x4 >>> 25);
      x3 = x1;
      x2 ^= x1;
      x3 <<= 7;
      x2 ^= x4;
      x0 ^= x4;
      x0 ^= x3;
      x2 = (x2 << 5) | (x2 >>> 27);
      x0 = (x0 << 22) | (x0 >>> 10);
      x2 ^= key.k64;
      x1 ^= key.k65;
      x0 ^= key.k66;
      x4 ^= key.k67;
      x4 ^= x2 ; x3 = x1 ;
      x1 &= x4 ; x3 ^= x0 ;
      x1 ^= x2 ; x2 |= x4 ;
      x2 ^= x3 ; x3 ^= x4 ;
      x4 ^= x0 ; x0 |= x1 ;
      x0 ^= x3 ; x3 ^= -1 ;
      x3 |= x1 ; x1 ^= x4 ;
      x1 ^= x3 ; x4 |= x2 ;
      x1 ^= x4 ; x3 ^= x4 ;
      x1 = (x1 << 13) | (x1 >>> 19);
      x3 ^= x1;
      x4 = x1 << 3;
      x0 = (x0 << 3) | (x0 >>> 29);
      x3 ^= x0;
      x2 ^= x0;
      x3 = (x3 << 1) | (x3 >>> 31);
      x2 ^= x4;
      x2 = (x2 << 7) | (x2 >>> 25);
      x4 = x3;
      x1 ^= x3;
      x4 <<= 7;
      x1 ^= x2;
      x0 ^= x2;
      x0 ^= x4;
      x1 = (x1 << 5) | (x1 >>> 27);
      x0 = (x0 << 22) | (x0 >>> 10);
      x1 ^= key.k68;
      x3 ^= key.k69;
      x0 ^= key.k70;
      x2 ^= key.k71;
      x3 =~ x3; x4 = x1;
      x1 ^= x3; x4 |= x3;
      x4 ^= x2; x2 &= x1;
      x0 ^= x4; x2 ^= x3;
      x2 |= x0; x1 ^= x4;
      x2 ^= x1; x3 &= x0;
      x1 |= x3; x3 ^= x4;
      x1 ^= x0; x4 |= x2;
      x1 ^= x4; x4 =~ x4;
      x3 ^= x2; x4 &= x0;
      x3 =~ x3; x4 ^= x1;
      x3 ^= x4;
      x2 = (x2 << 13) | (x2 >>> 19);
      x3 ^= x2;
      x4 = x2 << 3;
      x0 = (x0 << 3) | (x0 >>> 29);
      x3 ^= x0;
      x1 ^= x0;
      x3 = (x3 << 1) | (x3 >>> 31);
      x1 ^= x4;
      x1 = (x1 << 7) | (x1 >>> 25);
      x4 = x3;
      x2 ^= x3;
      x4 <<= 7;
      x2 ^= x1;
      x0 ^= x1;
      x0 ^= x4;
      x2 = (x2 << 5) | (x2 >>> 27);
      x0 = (x0 << 22) | (x0 >>> 10);
      x2 ^= key.k72;
      x3 ^= key.k73;
      x0 ^= key.k74;
      x1 ^= key.k75;
      x4 = x2 ;
      x2 = x2 & x0 ;
      x2 = x2 ^ x1 ;
      x0 = x0 ^ x3 ;
      x0 = x0 ^ x2 ;
      x1 = x1 | x4 ;
      x1 = x1 ^ x3 ;
      x4 = x4 ^ x0 ;
      x3 = x1 ;
      x1 = x1 | x4 ;
      x1 = x1 ^ x2 ;
      x2 = x2 & x3 ;
      x4 = x4 ^ x2 ;
      x3 = x3 ^ x1 ;
      x3 = x3 ^ x4 ;
      x4 = ~ x4 ;
      x0 = (x0 << 13) | (x0 >>> 19);
      x1 ^= x0;
      x2 = x0 << 3;
      x3 = (x3 << 3) | (x3 >>> 29);
      x1 ^= x3;
      x4 ^= x3;
      x1 = (x1 << 1) | (x1 >>> 31);
      x4 ^= x2;
      x4 = (x4 << 7) | (x4 >>> 25);
      x2 = x1;
      x0 ^= x1;
      x2 <<= 7;
      x0 ^= x4;
      x3 ^= x4;
      x3 ^= x2;
      x0 = (x0 << 5) | (x0 >>> 27);
      x3 = (x3 << 22) | (x3 >>> 10);
      x0 ^= key.k76;
      x1 ^= key.k77;
      x3 ^= key.k78;
      x4 ^= key.k79;
      x2 = x0 ;
      x0 = x0 | x4 ;
      x4 = x4 ^ x1 ;
      x1 = x1 & x2 ;
      x2 = x2 ^ x3 ;
      x3 = x3 ^ x4 ;
      x4 = x4 & x0 ;
      x2 = x2 | x1 ;
      x4 = x4 ^ x2 ;
      x0 = x0 ^ x1 ;
      x2 = x2 & x0 ;
      x1 = x1 ^ x4 ;
      x2 = x2 ^ x3 ;
      x1 = x1 | x0 ;
      x1 = x1 ^ x3 ;
      x0 = x0 ^ x4 ;
      x3 = x1 ;
      x1 = x1 | x4 ;
      x1 = x1 ^ x0 ;
      x1 = (x1 << 13) | (x1 >>> 19);
      x3 ^= x1;
      x0 = x1 << 3;
      x4 = (x4 << 3) | (x4 >>> 29);
      x3 ^= x4;
      x2 ^= x4;
      x3 = (x3 << 1) | (x3 >>> 31);
      x2 ^= x0;
      x2 = (x2 << 7) | (x2 >>> 25);
      x0 = x3;
      x1 ^= x3;
      x0 <<= 7;
      x1 ^= x2;
      x4 ^= x2;
      x4 ^= x0;
      x1 = (x1 << 5) | (x1 >>> 27);
      x4 = (x4 << 22) | (x4 >>> 10);
      x1 ^= key.k80;
      x3 ^= key.k81;
      x4 ^= key.k82;
      x2 ^= key.k83;
      x3 = x3 ^ x2 ;
      x2 = ~ x2 ;
      x4 = x4 ^ x2 ;
      x2 = x2 ^ x1 ;
      x0 = x3 ;
      x3 = x3 & x2 ;
      x3 = x3 ^ x4 ;
      x0 = x0 ^ x2 ;
      x1 = x1 ^ x0 ;
      x4 = x4 & x0 ;
      x4 = x4 ^ x1 ;
      x1 = x1 & x3 ;
      x2 = x2 ^ x1 ;
      x0 = x0 | x3 ;
      x0 = x0 ^ x1 ;
      x1 = x1 | x2 ;
      x1 = x1 ^ x4 ;
      x4 = x4 & x2 ;
      x1 = ~ x1 ;
      x0 = x0 ^ x4 ;
      x3 = (x3 << 13) | (x3 >>> 19);
      x0 ^= x3;
      x4 = x3 << 3;
      x1 = (x1 << 3) | (x1 >>> 29);
      x0 ^= x1;
      x2 ^= x1;
      x0 = (x0 << 1) | (x0 >>> 31);
      x2 ^= x4;
      x2 = (x2 << 7) | (x2 >>> 25);
      x4 = x0;
      x3 ^= x0;
      x4 <<= 7;
      x3 ^= x2;
      x1 ^= x2;
      x1 ^= x4;
      x3 = (x3 << 5) | (x3 >>> 27);
      x1 = (x1 << 22) | (x1 >>> 10);
      x3 ^= key.k84;
      x0 ^= key.k85;
      x1 ^= key.k86;
      x2 ^= key.k87;
      x3 = x3 ^ x0 ;
      x0 = x0 ^ x2 ;
      x2 = ~ x2 ;
      x4 = x0 ;
      x0 = x0 & x3 ;
      x1 = x1 ^ x2 ;
      x0 = x0 ^ x1 ;
      x1 = x1 | x4 ;
      x4 = x4 ^ x2 ;
      x2 = x2 & x0 ;
      x2 = x2 ^ x3 ;
      x4 = x4 ^ x0 ;
      x4 = x4 ^ x1 ;
      x1 = x1 ^ x3 ;
      x3 = x3 & x2 ;
      x1 = ~ x1 ;
      x3 = x3 ^ x4 ;
      x4 = x4 | x2 ;
      x1 = x1 ^ x4 ;
      x0 = (x0 << 13) | (x0 >>> 19);
      x2 ^= x0;
      x4 = x0 << 3;
      x3 = (x3 << 3) | (x3 >>> 29);
      x2 ^= x3;
      x1 ^= x3;
      x2 = (x2 << 1) | (x2 >>> 31);
      x1 ^= x4;
      x1 = (x1 << 7) | (x1 >>> 25);
      x4 = x2;
      x0 ^= x2;
      x4 <<= 7;
      x0 ^= x1;
      x3 ^= x1;
      x3 ^= x4;
      x0 = (x0 << 5) | (x0 >>> 27);
      x3 = (x3 << 22) | (x3 >>> 10);
      x0 ^= key.k88;
      x2 ^= key.k89;
      x3 ^= key.k90;
      x1 ^= key.k91;
      x3 = ~ x3 ;
      x4 = x1 ;
      x1 = x1 & x0 ;
      x0 = x0 ^ x4 ;
      x1 = x1 ^ x3 ;
      x3 = x3 | x4 ;
      x2 = x2 ^ x1 ;
      x3 = x3 ^ x0 ;
      x0 = x0 | x2 ;
      x3 = x3 ^ x2 ;
      x4 = x4 ^ x0 ;
      x0 = x0 | x1 ;
      x0 = x0 ^ x3 ;
      x4 = x4 ^ x1 ;
      x4 = x4 ^ x0 ;
      x1 = ~ x1 ;
      x3 = x3 & x4 ;
      x3 = x3 ^ x1 ;
      x0 = (x0 << 13) | (x0 >>> 19);
      x2 ^= x0;
      x1 = x0 << 3;
      x4 = (x4 << 3) | (x4 >>> 29);
      x2 ^= x4;
      x3 ^= x4;
      x2 = (x2 << 1) | (x2 >>> 31);
      x3 ^= x1;
      x3 = (x3 << 7) | (x3 >>> 25);
      x1 = x2;
      x0 ^= x2;
      x1 <<= 7;
      x0 ^= x3;
      x4 ^= x3;
      x4 ^= x1;
      x0 = (x0 << 5) | (x0 >>> 27);
      x4 = (x4 << 22) | (x4 >>> 10);
      x0 ^= key.k92;
      x2 ^= key.k93;
      x4 ^= key.k94;
      x3 ^= key.k95;
      x1 = x4 ;
      x4 = x4 & x2 ;
      x4 = x4 ^ x3 ;
      x3 = x3 & x2 ;
      x1 = x1 ^ x4 ;
      x4 = x4 ^ x2 ;
      x2 = x2 ^ x0 ;
      x0 = x0 | x1 ;
      x0 = x0 ^ x4 ;
      x3 = x3 ^ x2 ;
      x4 = x4 ^ x3 ;
      x3 = x3 & x0 ;
      x3 = x3 ^ x1 ;
      x1 = x1 ^ x4 ;
      x4 = x4 & x0 ;
      x1 = ~ x1 ;
      x4 = x4 ^ x1 ;
      x1 = x1 & x0 ;
      x2 = x2 ^ x3 ;
      x1 = x1 ^ x2 ;
      x4 = (x4 << 13) | (x4 >>> 19);
      x1 ^= x4;
      x2 = x4 << 3;
      x3 = (x3 << 3) | (x3 >>> 29);
      x1 ^= x3;
      x0 ^= x3;
      x1 = (x1 << 1) | (x1 >>> 31);
      x0 ^= x2;
      x0 = (x0 << 7) | (x0 >>> 25);
      x2 = x1;
      x4 ^= x1;
      x2 <<= 7;
      x4 ^= x0;
      x3 ^= x0;
      x3 ^= x2;
      x4 = (x4 << 5) | (x4 >>> 27);
      x3 = (x3 << 22) | (x3 >>> 10);
      x4 ^= key.k96;
      x1 ^= key.k97;
      x3 ^= key.k98;
      x0 ^= key.k99;
      x0 ^= x4 ; x2 = x1 ;
      x1 &= x0 ; x2 ^= x3 ;
      x1 ^= x4 ; x4 |= x0 ;
      x4 ^= x2 ; x2 ^= x0 ;
      x0 ^= x3 ; x3 |= x1 ;
      x3 ^= x2 ; x2 ^= -1 ;
      x2 |= x1 ; x1 ^= x0 ;
      x1 ^= x2 ; x0 |= x4 ;
      x1 ^= x0 ; x2 ^= x0 ;
      x1 = (x1 << 13) | (x1 >>> 19);
      x2 ^= x1;
      x0 = x1 << 3;
      x3 = (x3 << 3) | (x3 >>> 29);
      x2 ^= x3;
      x4 ^= x3;
      x2 = (x2 << 1) | (x2 >>> 31);
      x4 ^= x0;
      x4 = (x4 << 7) | (x4 >>> 25);
      x0 = x2;
      x1 ^= x2;
      x0 <<= 7;
      x1 ^= x4;
      x3 ^= x4;
      x3 ^= x0;
      x1 = (x1 << 5) | (x1 >>> 27);
      x3 = (x3 << 22) | (x3 >>> 10);
      x1 ^= key.k100;
      x2 ^= key.k101;
      x3 ^= key.k102;
      x4 ^= key.k103;
      x2 =~ x2; x0 = x1;
      x1 ^= x2; x0 |= x2;
      x0 ^= x4; x4 &= x1;
      x3 ^= x0; x4 ^= x2;
      x4 |= x3; x1 ^= x0;
      x4 ^= x1; x2 &= x3;
      x1 |= x2; x2 ^= x0;
      x1 ^= x3; x0 |= x4;
      x1 ^= x0; x0 =~ x0;
      x2 ^= x4; x0 &= x3;
      x2 =~ x2; x0 ^= x1;
      x2 ^= x0;
      x4 = (x4 << 13) | (x4 >>> 19);
      x2 ^= x4;
      x0 = x4 << 3;
      x3 = (x3 << 3) | (x3 >>> 29);
      x2 ^= x3;
      x1 ^= x3;
      x2 = (x2 << 1) | (x2 >>> 31);
      x1 ^= x0;
      x1 = (x1 << 7) | (x1 >>> 25);
      x0 = x2;
      x4 ^= x2;
      x0 <<= 7;
      x4 ^= x1;
      x3 ^= x1;
      x3 ^= x0;
      x4 = (x4 << 5) | (x4 >>> 27);
      x3 = (x3 << 22) | (x3 >>> 10);
      x4 ^= key.k104;
      x2 ^= key.k105;
      x3 ^= key.k106;
      x1 ^= key.k107;
      x0 = x4 ;
      x4 = x4 & x3 ;
      x4 = x4 ^ x1 ;
      x3 = x3 ^ x2 ;
      x3 = x3 ^ x4 ;
      x1 = x1 | x0 ;
      x1 = x1 ^ x2 ;
      x0 = x0 ^ x3 ;
      x2 = x1 ;
      x1 = x1 | x0 ;
      x1 = x1 ^ x4 ;
      x4 = x4 & x2 ;
      x0 = x0 ^ x4 ;
      x2 = x2 ^ x1 ;
      x2 = x2 ^ x0 ;
      x0 = ~ x0 ;
      x3 = (x3 << 13) | (x3 >>> 19);
      x1 ^= x3;
      x4 = x3 << 3;
      x2 = (x2 << 3) | (x2 >>> 29);
      x1 ^= x2;
      x0 ^= x2;
      x1 = (x1 << 1) | (x1 >>> 31);
      x0 ^= x4;
      x0 = (x0 << 7) | (x0 >>> 25);
      x4 = x1;
      x3 ^= x1;
      x4 <<= 7;
      x3 ^= x0;
      x2 ^= x0;
      x2 ^= x4;
      x3 = (x3 << 5) | (x3 >>> 27);
      x2 = (x2 << 22) | (x2 >>> 10);
      x3 ^= key.k108;
      x1 ^= key.k109;
      x2 ^= key.k110;
      x0 ^= key.k111;
      x4 = x3 ;
      x3 = x3 | x0 ;
      x0 = x0 ^ x1 ;
      x1 = x1 & x4 ;
      x4 = x4 ^ x2 ;
      x2 = x2 ^ x0 ;
      x0 = x0 & x3 ;
      x4 = x4 | x1 ;
      x0 = x0 ^ x4 ;
      x3 = x3 ^ x1 ;
      x4 = x4 & x3 ;
      x1 = x1 ^ x0 ;
      x4 = x4 ^ x2 ;
      x1 = x1 | x3 ;
      x1 = x1 ^ x2 ;
      x3 = x3 ^ x0 ;
      x2 = x1 ;
      x1 = x1 | x0 ;
      x1 = x1 ^ x3 ;
      x1 = (x1 << 13) | (x1 >>> 19);
      x2 ^= x1;
      x3 = x1 << 3;
      x0 = (x0 << 3) | (x0 >>> 29);
      x2 ^= x0;
      x4 ^= x0;
      x2 = (x2 << 1) | (x2 >>> 31);
      x4 ^= x3;
      x4 = (x4 << 7) | (x4 >>> 25);
      x3 = x2;
      x1 ^= x2;
      x3 <<= 7;
      x1 ^= x4;
      x0 ^= x4;
      x0 ^= x3;
      x1 = (x1 << 5) | (x1 >>> 27);
      x0 = (x0 << 22) | (x0 >>> 10);
      x1 ^= key.k112;
      x2 ^= key.k113;
      x0 ^= key.k114;
      x4 ^= key.k115;
      x2 = x2 ^ x4 ;
      x4 = ~ x4 ;
      x0 = x0 ^ x4 ;
      x4 = x4 ^ x1 ;
      x3 = x2 ;
      x2 = x2 & x4 ;
      x2 = x2 ^ x0 ;
      x3 = x3 ^ x4 ;
      x1 = x1 ^ x3 ;
      x0 = x0 & x3 ;
      x0 = x0 ^ x1 ;
      x1 = x1 & x2 ;
      x4 = x4 ^ x1 ;
      x3 = x3 | x2 ;
      x3 = x3 ^ x1 ;
      x1 = x1 | x4 ;
      x1 = x1 ^ x0 ;
      x0 = x0 & x4 ;
      x1 = ~ x1 ;
      x3 = x3 ^ x0 ;
      x2 = (x2 << 13) | (x2 >>> 19);
      x3 ^= x2;
      x0 = x2 << 3;
      x1 = (x1 << 3) | (x1 >>> 29);
      x3 ^= x1;
      x4 ^= x1;
      x3 = (x3 << 1) | (x3 >>> 31);
      x4 ^= x0;
      x4 = (x4 << 7) | (x4 >>> 25);
      x0 = x3;
      x2 ^= x3;
      x0 <<= 7;
      x2 ^= x4;
      x1 ^= x4;
      x1 ^= x0;
      x2 = (x2 << 5) | (x2 >>> 27);
      x1 = (x1 << 22) | (x1 >>> 10);
      x2 ^= key.k116;
      x3 ^= key.k117;
      x1 ^= key.k118;
      x4 ^= key.k119;
      x2 = x2 ^ x3 ;
      x3 = x3 ^ x4 ;
      x4 = ~ x4 ;
      x0 = x3 ;
      x3 = x3 & x2 ;
      x1 = x1 ^ x4 ;
      x3 = x3 ^ x1 ;
      x1 = x1 | x0 ;
      x0 = x0 ^ x4 ;
      x4 = x4 & x3 ;
      x4 = x4 ^ x2 ;
      x0 = x0 ^ x3 ;
      x0 = x0 ^ x1 ;
      x1 = x1 ^ x2 ;
      x2 = x2 & x4 ;
      x1 = ~ x1 ;
      x2 = x2 ^ x0 ;
      x0 = x0 | x4 ;
      x1 = x1 ^ x0 ;
      x3 = (x3 << 13) | (x3 >>> 19);
      x4 ^= x3;
      x0 = x3 << 3;
      x2 = (x2 << 3) | (x2 >>> 29);
      x4 ^= x2;
      x1 ^= x2;
      x4 = (x4 << 1) | (x4 >>> 31);
      x1 ^= x0;
      x1 = (x1 << 7) | (x1 >>> 25);
      x0 = x4;
      x3 ^= x4;
      x0 <<= 7;
      x3 ^= x1;
      x2 ^= x1;
      x2 ^= x0;
      x3 = (x3 << 5) | (x3 >>> 27);
      x2 = (x2 << 22) | (x2 >>> 10);
      x3 ^= key.k120;
      x4 ^= key.k121;
      x2 ^= key.k122;
      x1 ^= key.k123;
      x2 = ~ x2 ;
      x0 = x1 ;
      x1 = x1 & x3 ;
      x3 = x3 ^ x0 ;
      x1 = x1 ^ x2 ;
      x2 = x2 | x0 ;
      x4 = x4 ^ x1 ;
      x2 = x2 ^ x3 ;
      x3 = x3 | x4 ;
      x2 = x2 ^ x4 ;
      x0 = x0 ^ x3 ;
      x3 = x3 | x1 ;
      x3 = x3 ^ x2 ;
      x0 = x0 ^ x1 ;
      x0 = x0 ^ x3 ;
      x1 = ~ x1 ;
      x2 = x2 & x0 ;
      x2 = x2 ^ x1 ;
      x3 = (x3 << 13) | (x3 >>> 19);
      x4 ^= x3;
      x1 = x3 << 3;
      x0 = (x0 << 3) | (x0 >>> 29);
      x4 ^= x0;
      x2 ^= x0;
      x4 = (x4 << 1) | (x4 >>> 31);
      x2 ^= x1;
      x2 = (x2 << 7) | (x2 >>> 25);
      x1 = x4;
      x3 ^= x4;
      x1 <<= 7;
      x3 ^= x2;
      x0 ^= x2;
      x0 ^= x1;
      x3 = (x3 << 5) | (x3 >>> 27);
      x0 = (x0 << 22) | (x0 >>> 10);
      x3 ^= key.k124;
      x4 ^= key.k125;
      x0 ^= key.k126;
      x2 ^= key.k127;
      x1 = x0 ;
      x0 = x0 & x4 ;
      x0 = x0 ^ x2 ;
      x2 = x2 & x4 ;
      x1 = x1 ^ x0 ;
      x0 = x0 ^ x4 ;
      x4 = x4 ^ x3 ;
      x3 = x3 | x1 ;
      x3 = x3 ^ x0 ;
      x2 = x2 ^ x4 ;
      x0 = x0 ^ x2 ;
      x2 = x2 & x3 ;
      x2 = x2 ^ x1 ;
      x1 = x1 ^ x0 ;
      x0 = x0 & x3 ;
      x1 = ~ x1 ;
      x0 = x0 ^ x1 ;
      x1 = x1 & x3 ;
      x4 = x4 ^ x2 ;
      x1 = x1 ^ x4 ;
      x0 ^= key.k128;
      x1 ^= key.k129;
      x2 ^= key.k130;
      x3 ^= key.k131;

      out[o   ] = (byte) (x3 >>> 24);
      out[o+ 1] = (byte) (x3 >>> 16);
      out[o+ 2] = (byte) (x3 >>>  8);
      out[o+ 3] = (byte)  x3;
      out[o+ 4] = (byte) (x2 >>> 24);
      out[o+ 5] = (byte) (x2 >>> 16);
      out[o+ 6] = (byte) (x2 >>>  8);
      out[o+ 7] = (byte)  x2;
      out[o+ 8] = (byte) (x1 >>> 24);
      out[o+ 9] = (byte) (x1 >>> 16);
      out[o+10] = (byte) (x1 >>>  8);
      out[o+11] = (byte)  x1;
      out[o+12] = (byte) (x0 >>> 24);
      out[o+13] = (byte) (x0 >>> 16);
      out[o+14] = (byte) (x0 >>>  8);
      out[o+15] = (byte)  x0;
   }

   public void decrypt(byte[] in, int i, byte[] out, int o, Object K, int bs) {
      SerpentKey key = (SerpentKey) K;
      //int x0, x1, x2, x3, x4;
      x3 = (in[i   ] & 0xff) << 24 | (in[i+ 1] & 0xff) << 16 |
           (in[i+ 2] & 0xff) <<  8 | (in[i+ 3] & 0xff);
      x2 = (in[i+ 4] & 0xff) << 24 | (in[i+ 5] & 0xff) << 16 |
           (in[i+ 6] & 0xff) <<  8 | (in[i+ 7] & 0xff);
      x1 = (in[i+ 8] & 0xff) << 24 | (in[i+ 9] & 0xff) << 16 |
           (in[i+10] & 0xff) <<  8 | (in[i+11] & 0xff);
      x0 = (in[i+12] & 0xff) << 24 | (in[i+13] & 0xff) << 16 |
           (in[i+14] & 0xff) <<  8 | (in[i+15] & 0xff);

      x0 ^= key.k128;
      x1 ^= key.k129;
      x2 ^= key.k130;
      x3 ^= key.k131;
      x4 = x2; x2 ^= x0;
      x0 &= x3;
      x2 =~ x2; x4 |= x3;
      x3 ^= x1; x1 |= x0;
      x0 ^= x2; x2 &= x4;
      x1 ^= x2; x2 ^= x0;
      x0 |= x2; x3 &= x4;
      x0 ^= x3; x4 ^= x1;
      x3 ^= x4; x4 |= x0;
      x3 ^= x2; x4 ^= x2;
      x3 ^= key.k124;
      x0 ^= key.k125;
      x1 ^= key.k126;
      x4 ^= key.k127;
      x1 = (x1 >>> 22) | (x1 << 10);
      x3 = (x3 >>> 5) | (x3 << 27);
      x2 = x0;
      x1 ^= x4;
      x2 <<= 7;
      x3 ^= x4;
      x1 ^= x2;
      x3 ^= x0;
      x4 = (x4 >>> 7) | (x4 << 25);
      x0 = (x0 >>> 1) | (x0 << 31);
      x0 ^= x3;
      x2 = x3 << 3;
      x4 ^= x2;
      x3 = (x3 >>> 13) | (x3 << 19);
      x0 ^= x1;
      x4 ^= x1;
      x1 = (x1 >>> 3) | (x1 << 29);
      x3 ^= x1; x2 = x1;
      x1 &= x3; x2 ^= x4;
      x1 =~ x1; x4 ^= x0;
      x1 ^= x4; x2 |= x3;
      x3 ^= x1; x4 ^= x2;
      x2 ^= x0; x0 &= x4;
      x0 ^= x3; x3 ^= x4;
      x3 |= x1; x4 ^= x0;
      x2 ^= x3;
      x0 ^= key.k120;
      x1 ^= key.k121;
      x2 ^= key.k122;
      x4 ^= key.k123;
      x2 = (x2 >>> 22) | (x2 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x3 = x1;
      x2 ^= x4;
      x3 <<= 7;
      x0 ^= x4;
      x2 ^= x3;
      x0 ^= x1;
      x4 = (x4 >>> 7) | (x4 << 25);
      x1 = (x1 >>> 1) | (x1 << 31);
      x1 ^= x0;
      x3 = x0 << 3;
      x4 ^= x3;
      x0 = (x0 >>> 13) | (x0 << 19);
      x1 ^= x2;
      x4 ^= x2;
      x2 = (x2 >>> 3) | (x2 << 29);
      x1 =~ x1; x3 = x4;
      x2 ^= x1; x4 |= x0;
      x4 ^= x2; x2 |= x1;
      x2 &= x0; x3 ^= x4;
      x2 ^= x3; x3 |= x0;
      x3 ^= x1; x1 &= x2;
      x1 ^= x4; x3 ^= x2;
      x4 &= x3; x3 ^= x1;
      x4 ^= x0;
      x4 ^= x3; x3 =~ x3;
      x1 ^= key.k116;
      x3 ^= key.k117;
      x4 ^= key.k118;
      x2 ^= key.k119;
      x4 = (x4 >>> 22) | (x4 << 10);
      x1 = (x1 >>> 5) | (x1 << 27);
      x0 = x3;
      x4 ^= x2;
      x0 <<= 7;
      x1 ^= x2;
      x4 ^= x0;
      x1 ^= x3;
      x2 = (x2 >>> 7) | (x2 << 25);
      x3 = (x3 >>> 1) | (x3 << 31);
      x3 ^= x1;
      x0 = x1 << 3;
      x2 ^= x0;
      x1 = (x1 >>> 13) | (x1 << 19);
      x3 ^= x4;
      x2 ^= x4;
      x4 = (x4 >>> 3) | (x4 << 29);
      x0 = x4; x4 &= x2;
      x4 ^= x3; x3 |= x2;
      x3 &= x1; x0 ^= x4;
      x0 ^= x3; x3 &= x4;
      x1 =~ x1; x2 ^= x0;
      x3 ^= x2; x2 &= x1;
      x2 ^= x4; x1 ^= x3;
      x4 &= x1; x2 ^= x1;
      x4 ^= x0;
      x4 |= x2; x2 ^= x1;
      x4 ^= x3;
      x1 ^= key.k112;
      x2 ^= key.k113;
      x4 ^= key.k114;
      x0 ^= key.k115;
      x4 = (x4 >>> 22) | (x4 << 10);
      x1 = (x1 >>> 5) | (x1 << 27);
      x3 = x2;
      x4 ^= x0;
      x3 <<= 7;
      x1 ^= x0;
      x4 ^= x3;
      x1 ^= x2;
      x0 = (x0 >>> 7) | (x0 << 25);
      x2 = (x2 >>> 1) | (x2 << 31);
      x2 ^= x1;
      x3 = x1 << 3;
      x0 ^= x3;
      x1 = (x1 >>> 13) | (x1 << 19);
      x2 ^= x4;
      x0 ^= x4;
      x4 = (x4 >>> 3) | (x4 << 29);
      x3 = x4; x4 ^= x2;
      x2 &= x4;
      x2 ^= x1; x1 &= x3;
      x3 ^= x0; x0 |= x2;
      x0 ^= x4; x1 ^= x3;
      x4 ^= x1; x1 |= x0;
      x1 ^= x2; x3 ^= x4;
      x4 &= x0; x2 |= x0;
      x2 ^= x4; x3 ^= x1;
      x4 ^= x3;
      x0 ^= key.k108;
      x1 ^= key.k109;
      x4 ^= key.k110;
      x2 ^= key.k111;
      x4 = (x4 >>> 22) | (x4 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x3 = x1;
      x4 ^= x2;
      x3 <<= 7;
      x0 ^= x2;
      x4 ^= x3;
      x0 ^= x1;
      x2 = (x2 >>> 7) | (x2 << 25);
      x1 = (x1 >>> 1) | (x1 << 31);
      x1 ^= x0;
      x3 = x0 << 3;
      x2 ^= x3;
      x0 = (x0 >>> 13) | (x0 << 19);
      x1 ^= x4;
      x2 ^= x4;
      x4 = (x4 >>> 3) | (x4 << 29);
      x4 ^= x2; x2 ^= x0;
      x3 = x2; x2 &= x4;
      x2 ^= x1; x1 |= x4;
      x1 ^= x3; x3 &= x2;
      x4 ^= x2; x3 &= x0;
      x3 ^= x4; x4 &= x1;
      x4 |= x0; x2 =~ x2;
      x4 ^= x2; x0 ^= x2;
      x0 &= x1; x2 ^= x3;
      x2 ^= x0;
      x1 ^= key.k104;
      x3 ^= key.k105;
      x4 ^= key.k106;
      x2 ^= key.k107;
      x4 = (x4 >>> 22) | (x4 << 10);
      x1 = (x1 >>> 5) | (x1 << 27);
      x0 = x3;
      x4 ^= x2;
      x0 <<= 7;
      x1 ^= x2;
      x4 ^= x0;
      x1 ^= x3;
      x2 = (x2 >>> 7) | (x2 << 25);
      x3 = (x3 >>> 1) | (x3 << 31);
      x3 ^= x1;
      x0 = x1 << 3;
      x2 ^= x0;
      x1 = (x1 >>> 13) | (x1 << 19);
      x3 ^= x4;
      x2 ^= x4;
      x4 = (x4 >>> 3) | (x4 << 29);
      x0 = x3; x3 ^= x2;
      x2 &= x3; x0 ^= x4;
      x2 ^= x1; x1 |= x3;
      x4 ^= x2; x1 ^= x0;
      x1 |= x4; x3 ^= x2;
      x1 ^= x3; x3 |= x2;
      x3 ^= x1; x0 =~ x0;
      x0 ^= x3; x3 |= x1;
      x3 ^= x1;
      x3 |= x0;
      x2 ^= x3;
      x0 ^= key.k100;
      x1 ^= key.k101;
      x2 ^= key.k102;
      x4 ^= key.k103;
      x2 = (x2 >>> 22) | (x2 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x3 = x1;
      x2 ^= x4;
      x3 <<= 7;
      x0 ^= x4;
      x2 ^= x3;
      x0 ^= x1;
      x4 = (x4 >>> 7) | (x4 << 25);
      x1 = (x1 >>> 1) | (x1 << 31);
      x1 ^= x0;
      x3 = x0 << 3;
      x4 ^= x3;
      x0 = (x0 >>> 13) | (x0 << 19);
      x1 ^= x2;
      x4 ^= x2;
      x2 = (x2 >>> 3) | (x2 << 29);
      x2 =~ x2; x3 = x1;
      x1 |= x0; x3 =~ x3;
      x1 ^= x2; x2 |= x3;
      x1 ^= x4; x0 ^= x3;
      x2 ^= x0; x0 &= x4;
      x3 ^= x0; x0 |= x1;
      x0 ^= x2; x4 ^= x3;
      x2 ^= x1; x4 ^= x0;
      x4 ^= x1;
      x2 &= x4;
      x3 ^= x2;
      x0 ^= key.k96;
      x3 ^= key.k97;
      x1 ^= key.k98;
      x4 ^= key.k99;
      x1 = (x1 >>> 22) | (x1 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x2 = x3;
      x1 ^= x4;
      x2 <<= 7;
      x0 ^= x4;
      x1 ^= x2;
      x0 ^= x3;
      x4 = (x4 >>> 7) | (x4 << 25);
      x3 = (x3 >>> 1) | (x3 << 31);
      x3 ^= x0;
      x2 = x0 << 3;
      x4 ^= x2;
      x0 = (x0 >>> 13) | (x0 << 19);
      x3 ^= x1;
      x4 ^= x1;
      x1 = (x1 >>> 3) | (x1 << 29);
      x2 = x1; x1 ^= x0;
      x0 &= x4;
      x1 =~ x1; x2 |= x4;
      x4 ^= x3; x3 |= x0;
      x0 ^= x1; x1 &= x2;
      x3 ^= x1; x1 ^= x0;
      x0 |= x1; x4 &= x2;
      x0 ^= x4; x2 ^= x3;
      x4 ^= x2; x2 |= x0;
      x4 ^= x1; x2 ^= x1;
      x4 ^= key.k92;
      x0 ^= key.k93;
      x3 ^= key.k94;
      x2 ^= key.k95;
      x3 = (x3 >>> 22) | (x3 << 10);
      x4 = (x4 >>> 5) | (x4 << 27);
      x1 = x0;
      x3 ^= x2;
      x1 <<= 7;
      x4 ^= x2;
      x3 ^= x1;
      x4 ^= x0;
      x2 = (x2 >>> 7) | (x2 << 25);
      x0 = (x0 >>> 1) | (x0 << 31);
      x0 ^= x4;
      x1 = x4 << 3;
      x2 ^= x1;
      x4 = (x4 >>> 13) | (x4 << 19);
      x0 ^= x3;
      x2 ^= x3;
      x3 = (x3 >>> 3) | (x3 << 29);
      x4 ^= x3; x1 = x3;
      x3 &= x4; x1 ^= x2;
      x3 =~ x3; x2 ^= x0;
      x3 ^= x2; x1 |= x4;
      x4 ^= x3; x2 ^= x1;
      x1 ^= x0; x0 &= x2;
      x0 ^= x4; x4 ^= x2;
      x4 |= x3; x2 ^= x0;
      x1 ^= x4;
      x0 ^= key.k88;
      x3 ^= key.k89;
      x1 ^= key.k90;
      x2 ^= key.k91;
      x1 = (x1 >>> 22) | (x1 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x4 = x3;
      x1 ^= x2;
      x4 <<= 7;
      x0 ^= x2;
      x1 ^= x4;
      x0 ^= x3;
      x2 = (x2 >>> 7) | (x2 << 25);
      x3 = (x3 >>> 1) | (x3 << 31);
      x3 ^= x0;
      x4 = x0 << 3;
      x2 ^= x4;
      x0 = (x0 >>> 13) | (x0 << 19);
      x3 ^= x1;
      x2 ^= x1;
      x1 = (x1 >>> 3) | (x1 << 29);
      x3 =~ x3; x4 = x2;
      x1 ^= x3; x2 |= x0;
      x2 ^= x1; x1 |= x3;
      x1 &= x0; x4 ^= x2;
      x1 ^= x4; x4 |= x0;
      x4 ^= x3; x3 &= x1;
      x3 ^= x2; x4 ^= x1;
      x2 &= x4; x4 ^= x3;
      x2 ^= x0;
      x2 ^= x4; x4 =~ x4;
      x3 ^= key.k84;
      x4 ^= key.k85;
      x2 ^= key.k86;
      x1 ^= key.k87;
      x2 = (x2 >>> 22) | (x2 << 10);
      x3 = (x3 >>> 5) | (x3 << 27);
      x0 = x4;
      x2 ^= x1;
      x0 <<= 7;
      x3 ^= x1;
      x2 ^= x0;
      x3 ^= x4;
      x1 = (x1 >>> 7) | (x1 << 25);
      x4 = (x4 >>> 1) | (x4 << 31);
      x4 ^= x3;
      x0 = x3 << 3;
      x1 ^= x0;
      x3 = (x3 >>> 13) | (x3 << 19);
      x4 ^= x2;
      x1 ^= x2;
      x2 = (x2 >>> 3) | (x2 << 29);
      x0 = x2; x2 &= x1;
      x2 ^= x4; x4 |= x1;
      x4 &= x3; x0 ^= x2;
      x0 ^= x4; x4 &= x2;
      x3 =~ x3; x1 ^= x0;
      x4 ^= x1; x1 &= x3;
      x1 ^= x2; x3 ^= x4;
      x2 &= x3; x1 ^= x3;
      x2 ^= x0;
      x2 |= x1; x1 ^= x3;
      x2 ^= x4;
      x3 ^= key.k80;
      x1 ^= key.k81;
      x2 ^= key.k82;
      x0 ^= key.k83;
      x2 = (x2 >>> 22) | (x2 << 10);
      x3 = (x3 >>> 5) | (x3 << 27);
      x4 = x1;
      x2 ^= x0;
      x4 <<= 7;
      x3 ^= x0;
      x2 ^= x4;
      x3 ^= x1;
      x0 = (x0 >>> 7) | (x0 << 25);
      x1 = (x1 >>> 1) | (x1 << 31);
      x1 ^= x3;
      x4 = x3 << 3;
      x0 ^= x4;
      x3 = (x3 >>> 13) | (x3 << 19);
      x1 ^= x2;
      x0 ^= x2;
      x2 = (x2 >>> 3) | (x2 << 29);
      x4 = x2; x2 ^= x1;
      x1 &= x2;
      x1 ^= x3; x3 &= x4;
      x4 ^= x0; x0 |= x1;
      x0 ^= x2; x3 ^= x4;
      x2 ^= x3; x3 |= x0;
      x3 ^= x1; x4 ^= x2;
      x2 &= x0; x1 |= x0;
      x1 ^= x2; x4 ^= x3;
      x2 ^= x4;
      x0 ^= key.k76;
      x3 ^= key.k77;
      x2 ^= key.k78;
      x1 ^= key.k79;
      x2 = (x2 >>> 22) | (x2 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x4 = x3;
      x2 ^= x1;
      x4 <<= 7;
      x0 ^= x1;
      x2 ^= x4;
      x0 ^= x3;
      x1 = (x1 >>> 7) | (x1 << 25);
      x3 = (x3 >>> 1) | (x3 << 31);
      x3 ^= x0;
      x4 = x0 << 3;
      x1 ^= x4;
      x0 = (x0 >>> 13) | (x0 << 19);
      x3 ^= x2;
      x1 ^= x2;
      x2 = (x2 >>> 3) | (x2 << 29);
      x2 ^= x1; x1 ^= x0;
      x4 = x1; x1 &= x2;
      x1 ^= x3; x3 |= x2;
      x3 ^= x4; x4 &= x1;
      x2 ^= x1; x4 &= x0;
      x4 ^= x2; x2 &= x3;
      x2 |= x0; x1 =~ x1;
      x2 ^= x1; x0 ^= x1;
      x0 &= x3; x1 ^= x4;
      x1 ^= x0;
      x3 ^= key.k72;
      x4 ^= key.k73;
      x2 ^= key.k74;
      x1 ^= key.k75;
      x2 = (x2 >>> 22) | (x2 << 10);
      x3 = (x3 >>> 5) | (x3 << 27);
      x0 = x4;
      x2 ^= x1;
      x0 <<= 7;
      x3 ^= x1;
      x2 ^= x0;
      x3 ^= x4;
      x1 = (x1 >>> 7) | (x1 << 25);
      x4 = (x4 >>> 1) | (x4 << 31);
      x4 ^= x3;
      x0 = x3 << 3;
      x1 ^= x0;
      x3 = (x3 >>> 13) | (x3 << 19);
      x4 ^= x2;
      x1 ^= x2;
      x2 = (x2 >>> 3) | (x2 << 29);
      x0 = x4; x4 ^= x1;
      x1 &= x4; x0 ^= x2;
      x1 ^= x3; x3 |= x4;
      x2 ^= x1; x3 ^= x0;
      x3 |= x2; x4 ^= x1;
      x3 ^= x4; x4 |= x1;
      x4 ^= x3; x0 =~ x0;
      x0 ^= x4; x4 |= x3;
      x4 ^= x3;
      x4 |= x0;
      x1 ^= x4;
      x0 ^= key.k68;
      x3 ^= key.k69;
      x1 ^= key.k70;
      x2 ^= key.k71;
      x1 = (x1 >>> 22) | (x1 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x4 = x3;
      x1 ^= x2;
      x4 <<= 7;
      x0 ^= x2;
      x1 ^= x4;
      x0 ^= x3;
      x2 = (x2 >>> 7) | (x2 << 25);
      x3 = (x3 >>> 1) | (x3 << 31);
      x3 ^= x0;
      x4 = x0 << 3;
      x2 ^= x4;
      x0 = (x0 >>> 13) | (x0 << 19);
      x3 ^= x1;
      x2 ^= x1;
      x1 = (x1 >>> 3) | (x1 << 29);
      x1 =~ x1; x4 = x3;
      x3 |= x0; x4 =~ x4;
      x3 ^= x1; x1 |= x4;
      x3 ^= x2; x0 ^= x4;
      x1 ^= x0; x0 &= x2;
      x4 ^= x0; x0 |= x3;
      x0 ^= x1; x2 ^= x4;
      x1 ^= x3; x2 ^= x0;
      x2 ^= x3;
      x1 &= x2;
      x4 ^= x1;
      x0 ^= key.k64;
      x4 ^= key.k65;
      x3 ^= key.k66;
      x2 ^= key.k67;
      x3 = (x3 >>> 22) | (x3 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x1 = x4;
      x3 ^= x2;
      x1 <<= 7;
      x0 ^= x2;
      x3 ^= x1;
      x0 ^= x4;
      x2 = (x2 >>> 7) | (x2 << 25);
      x4 = (x4 >>> 1) | (x4 << 31);
      x4 ^= x0;
      x1 = x0 << 3;
      x2 ^= x1;
      x0 = (x0 >>> 13) | (x0 << 19);
      x4 ^= x3;
      x2 ^= x3;
      x3 = (x3 >>> 3) | (x3 << 29);
      x1 = x3; x3 ^= x0;
      x0 &= x2;
      x3 =~ x3; x1 |= x2;
      x2 ^= x4; x4 |= x0;
      x0 ^= x3; x3 &= x1;
      x4 ^= x3; x3 ^= x0;
      x0 |= x3; x2 &= x1;
      x0 ^= x2; x1 ^= x4;
      x2 ^= x1; x1 |= x0;
      x2 ^= x3; x1 ^= x3;
      x2 ^= key.k60;
      x0 ^= key.k61;
      x4 ^= key.k62;
      x1 ^= key.k63;
      x4 = (x4 >>> 22) | (x4 << 10);
      x2 = (x2 >>> 5) | (x2 << 27);
      x3 = x0;
      x4 ^= x1;
      x3 <<= 7;
      x2 ^= x1;
      x4 ^= x3;
      x2 ^= x0;
      x1 = (x1 >>> 7) | (x1 << 25);
      x0 = (x0 >>> 1) | (x0 << 31);
      x0 ^= x2;
      x3 = x2 << 3;
      x1 ^= x3;
      x2 = (x2 >>> 13) | (x2 << 19);
      x0 ^= x4;
      x1 ^= x4;
      x4 = (x4 >>> 3) | (x4 << 29);
      x2 ^= x4; x3 = x4;
      x4 &= x2; x3 ^= x1;
      x4 =~ x4; x1 ^= x0;
      x4 ^= x1; x3 |= x2;
      x2 ^= x4; x1 ^= x3;
      x3 ^= x0; x0 &= x1;
      x0 ^= x2; x2 ^= x1;
      x2 |= x4; x1 ^= x0;
      x3 ^= x2;
      x0 ^= key.k56;
      x4 ^= key.k57;
      x3 ^= key.k58;
      x1 ^= key.k59;
      x3 = (x3 >>> 22) | (x3 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x2 = x4;
      x3 ^= x1;
      x2 <<= 7;
      x0 ^= x1;
      x3 ^= x2;
      x0 ^= x4;
      x1 = (x1 >>> 7) | (x1 << 25);
      x4 = (x4 >>> 1) | (x4 << 31);
      x4 ^= x0;
      x2 = x0 << 3;
      x1 ^= x2;
      x0 = (x0 >>> 13) | (x0 << 19);
      x4 ^= x3;
      x1 ^= x3;
      x3 = (x3 >>> 3) | (x3 << 29);
      x4 =~ x4; x2 = x1;
      x3 ^= x4; x1 |= x0;
      x1 ^= x3; x3 |= x4;
      x3 &= x0; x2 ^= x1;
      x3 ^= x2; x2 |= x0;
      x2 ^= x4; x4 &= x3;
      x4 ^= x1; x2 ^= x3;
      x1 &= x2; x2 ^= x4;
      x1 ^= x0;
      x1 ^= x2; x2 =~ x2;
      x4 ^= key.k52;
      x2 ^= key.k53;
      x1 ^= key.k54;
      x3 ^= key.k55;
      x1 = (x1 >>> 22) | (x1 << 10);
      x4 = (x4 >>> 5) | (x4 << 27);
      x0 = x2;
      x1 ^= x3;
      x0 <<= 7;
      x4 ^= x3;
      x1 ^= x0;
      x4 ^= x2;
      x3 = (x3 >>> 7) | (x3 << 25);
      x2 = (x2 >>> 1) | (x2 << 31);
      x2 ^= x4;
      x0 = x4 << 3;
      x3 ^= x0;
      x4 = (x4 >>> 13) | (x4 << 19);
      x2 ^= x1;
      x3 ^= x1;
      x1 = (x1 >>> 3) | (x1 << 29);
      x0 = x1; x1 &= x3;
      x1 ^= x2; x2 |= x3;
      x2 &= x4; x0 ^= x1;
      x0 ^= x2; x2 &= x1;
      x4 =~ x4; x3 ^= x0;
      x2 ^= x3; x3 &= x4;
      x3 ^= x1; x4 ^= x2;
      x1 &= x4; x3 ^= x4;
      x1 ^= x0;
      x1 |= x3; x3 ^= x4;
      x1 ^= x2;
      x4 ^= key.k48;
      x3 ^= key.k49;
      x1 ^= key.k50;
      x0 ^= key.k51;
      x1 = (x1 >>> 22) | (x1 << 10);
      x4 = (x4 >>> 5) | (x4 << 27);
      x2 = x3;
      x1 ^= x0;
      x2 <<= 7;
      x4 ^= x0;
      x1 ^= x2;
      x4 ^= x3;
      x0 = (x0 >>> 7) | (x0 << 25);
      x3 = (x3 >>> 1) | (x3 << 31);
      x3 ^= x4;
      x2 = x4 << 3;
      x0 ^= x2;
      x4 = (x4 >>> 13) | (x4 << 19);
      x3 ^= x1;
      x0 ^= x1;
      x1 = (x1 >>> 3) | (x1 << 29);
      x2 = x1; x1 ^= x3;
      x3 &= x1;
      x3 ^= x4; x4 &= x2;
      x2 ^= x0; x0 |= x3;
      x0 ^= x1; x4 ^= x2;
      x1 ^= x4; x4 |= x0;
      x4 ^= x3; x2 ^= x1;
      x1 &= x0; x3 |= x0;
      x3 ^= x1; x2 ^= x4;
      x1 ^= x2;
      x0 ^= key.k44;
      x4 ^= key.k45;
      x1 ^= key.k46;
      x3 ^= key.k47;
      x1 = (x1 >>> 22) | (x1 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x2 = x4;
      x1 ^= x3;
      x2 <<= 7;
      x0 ^= x3;
      x1 ^= x2;
      x0 ^= x4;
      x3 = (x3 >>> 7) | (x3 << 25);
      x4 = (x4 >>> 1) | (x4 << 31);
      x4 ^= x0;
      x2 = x0 << 3;
      x3 ^= x2;
      x0 = (x0 >>> 13) | (x0 << 19);
      x4 ^= x1;
      x3 ^= x1;
      x1 = (x1 >>> 3) | (x1 << 29);
      x1 ^= x3; x3 ^= x0;
      x2 = x3; x3 &= x1;
      x3 ^= x4; x4 |= x1;
      x4 ^= x2; x2 &= x3;
      x1 ^= x3; x2 &= x0;
      x2 ^= x1; x1 &= x4;
      x1 |= x0; x3 =~ x3;
      x1 ^= x3; x0 ^= x3;
      x0 &= x4; x3 ^= x2;
      x3 ^= x0;
      x4 ^= key.k40;
      x2 ^= key.k41;
      x1 ^= key.k42;
      x3 ^= key.k43;
      x1 = (x1 >>> 22) | (x1 << 10);
      x4 = (x4 >>> 5) | (x4 << 27);
      x0 = x2;
      x1 ^= x3;
      x0 <<= 7;
      x4 ^= x3;
      x1 ^= x0;
      x4 ^= x2;
      x3 = (x3 >>> 7) | (x3 << 25);
      x2 = (x2 >>> 1) | (x2 << 31);
      x2 ^= x4;
      x0 = x4 << 3;
      x3 ^= x0;
      x4 = (x4 >>> 13) | (x4 << 19);
      x2 ^= x1;
      x3 ^= x1;
      x1 = (x1 >>> 3) | (x1 << 29);
      x0 = x2; x2 ^= x3;
      x3 &= x2; x0 ^= x1;
      x3 ^= x4; x4 |= x2;
      x1 ^= x3; x4 ^= x0;
      x4 |= x1; x2 ^= x3;
      x4 ^= x2; x2 |= x3;
      x2 ^= x4; x0 =~ x0;
      x0 ^= x2; x2 |= x4;
      x2 ^= x4;
      x2 |= x0;
      x3 ^= x2;
      x0 ^= key.k36;
      x4 ^= key.k37;
      x3 ^= key.k38;
      x1 ^= key.k39;
      x3 = (x3 >>> 22) | (x3 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x2 = x4;
      x3 ^= x1;
      x2 <<= 7;
      x0 ^= x1;
      x3 ^= x2;
      x0 ^= x4;
      x1 = (x1 >>> 7) | (x1 << 25);
      x4 = (x4 >>> 1) | (x4 << 31);
      x4 ^= x0;
      x2 = x0 << 3;
      x1 ^= x2;
      x0 = (x0 >>> 13) | (x0 << 19);
      x4 ^= x3;
      x1 ^= x3;
      x3 = (x3 >>> 3) | (x3 << 29);
      x3 =~ x3; x2 = x4;
      x4 |= x0; x2 =~ x2;
      x4 ^= x3; x3 |= x2;
      x4 ^= x1; x0 ^= x2;
      x3 ^= x0; x0 &= x1;
      x2 ^= x0; x0 |= x4;
      x0 ^= x3; x1 ^= x2;
      x3 ^= x4; x1 ^= x0;
      x1 ^= x4;
      x3 &= x1;
      x2 ^= x3;
      x0 ^= key.k32;
      x2 ^= key.k33;
      x4 ^= key.k34;
      x1 ^= key.k35;
      x4 = (x4 >>> 22) | (x4 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x3 = x2;
      x4 ^= x1;
      x3 <<= 7;
      x0 ^= x1;
      x4 ^= x3;
      x0 ^= x2;
      x1 = (x1 >>> 7) | (x1 << 25);
      x2 = (x2 >>> 1) | (x2 << 31);
      x2 ^= x0;
      x3 = x0 << 3;
      x1 ^= x3;
      x0 = (x0 >>> 13) | (x0 << 19);
      x2 ^= x4;
      x1 ^= x4;
      x4 = (x4 >>> 3) | (x4 << 29);
      x3 = x4; x4 ^= x0;
      x0 &= x1;
      x4 =~ x4; x3 |= x1;
      x1 ^= x2; x2 |= x0;
      x0 ^= x4; x4 &= x3;
      x2 ^= x4; x4 ^= x0;
      x0 |= x4; x1 &= x3;
      x0 ^= x1; x3 ^= x2;
      x1 ^= x3; x3 |= x0;
      x1 ^= x4; x3 ^= x4;
      x1 ^= key.k28;
      x0 ^= key.k29;
      x2 ^= key.k30;
      x3 ^= key.k31;
      x2 = (x2 >>> 22) | (x2 << 10);
      x1 = (x1 >>> 5) | (x1 << 27);
      x4 = x0;
      x2 ^= x3;
      x4 <<= 7;
      x1 ^= x3;
      x2 ^= x4;
      x1 ^= x0;
      x3 = (x3 >>> 7) | (x3 << 25);
      x0 = (x0 >>> 1) | (x0 << 31);
      x0 ^= x1;
      x4 = x1 << 3;
      x3 ^= x4;
      x1 = (x1 >>> 13) | (x1 << 19);
      x0 ^= x2;
      x3 ^= x2;
      x2 = (x2 >>> 3) | (x2 << 29);
      x1 ^= x2; x4 = x2;
      x2 &= x1; x4 ^= x3;
      x2 =~ x2; x3 ^= x0;
      x2 ^= x3; x4 |= x1;
      x1 ^= x2; x3 ^= x4;
      x4 ^= x0; x0 &= x3;
      x0 ^= x1; x1 ^= x3;
      x1 |= x2; x3 ^= x0;
      x4 ^= x1;
      x0 ^= key.k24;
      x2 ^= key.k25;
      x4 ^= key.k26;
      x3 ^= key.k27;
      x4 = (x4 >>> 22) | (x4 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x1 = x2;
      x4 ^= x3;
      x1 <<= 7;
      x0 ^= x3;
      x4 ^= x1;
      x0 ^= x2;
      x3 = (x3 >>> 7) | (x3 << 25);
      x2 = (x2 >>> 1) | (x2 << 31);
      x2 ^= x0;
      x1 = x0 << 3;
      x3 ^= x1;
      x0 = (x0 >>> 13) | (x0 << 19);
      x2 ^= x4;
      x3 ^= x4;
      x4 = (x4 >>> 3) | (x4 << 29);
      x2 =~ x2; x1 = x3;
      x4 ^= x2; x3 |= x0;
      x3 ^= x4; x4 |= x2;
      x4 &= x0; x1 ^= x3;
      x4 ^= x1; x1 |= x0;
      x1 ^= x2; x2 &= x4;
      x2 ^= x3; x1 ^= x4;
      x3 &= x1; x1 ^= x2;
      x3 ^= x0;
      x3 ^= x1; x1 =~ x1;
      x2 ^= key.k20;
      x1 ^= key.k21;
      x3 ^= key.k22;
      x4 ^= key.k23;
      x3 = (x3 >>> 22) | (x3 << 10);
      x2 = (x2 >>> 5) | (x2 << 27);
      x0 = x1;
      x3 ^= x4;
      x0 <<= 7;
      x2 ^= x4;
      x3 ^= x0;
      x2 ^= x1;
      x4 = (x4 >>> 7) | (x4 << 25);
      x1 = (x1 >>> 1) | (x1 << 31);
      x1 ^= x2;
      x0 = x2 << 3;
      x4 ^= x0;
      x2 = (x2 >>> 13) | (x2 << 19);
      x1 ^= x3;
      x4 ^= x3;
      x3 = (x3 >>> 3) | (x3 << 29);
      x0 = x3; x3 &= x4;
      x3 ^= x1; x1 |= x4;
      x1 &= x2; x0 ^= x3;
      x0 ^= x1; x1 &= x3;
      x2 =~ x2; x4 ^= x0;
      x1 ^= x4; x4 &= x2;
      x4 ^= x3; x2 ^= x1;
      x3 &= x2; x4 ^= x2;
      x3 ^= x0;
      x3 |= x4; x4 ^= x2;
      x3 ^= x1;
      x2 ^= key.k16;
      x4 ^= key.k17;
      x3 ^= key.k18;
      x0 ^= key.k19;
      x3 = (x3 >>> 22) | (x3 << 10);
      x2 = (x2 >>> 5) | (x2 << 27);
      x1 = x4;
      x3 ^= x0;
      x1 <<= 7;
      x2 ^= x0;
      x3 ^= x1;
      x2 ^= x4;
      x0 = (x0 >>> 7) | (x0 << 25);
      x4 = (x4 >>> 1) | (x4 << 31);
      x4 ^= x2;
      x1 = x2 << 3;
      x0 ^= x1;
      x2 = (x2 >>> 13) | (x2 << 19);
      x4 ^= x3;
      x0 ^= x3;
      x3 = (x3 >>> 3) | (x3 << 29);
      x1 = x3; x3 ^= x4;
      x4 &= x3;
      x4 ^= x2; x2 &= x1;
      x1 ^= x0; x0 |= x4;
      x0 ^= x3; x2 ^= x1;
      x3 ^= x2; x2 |= x0;
      x2 ^= x4; x1 ^= x3;
      x3 &= x0; x4 |= x0;
      x4 ^= x3; x1 ^= x2;
      x3 ^= x1;
      x0 ^= key.k12;
      x2 ^= key.k13;
      x3 ^= key.k14;
      x4 ^= key.k15;
      x3 = (x3 >>> 22) | (x3 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x1 = x2;
      x3 ^= x4;
      x1 <<= 7;
      x0 ^= x4;
      x3 ^= x1;
      x0 ^= x2;
      x4 = (x4 >>> 7) | (x4 << 25);
      x2 = (x2 >>> 1) | (x2 << 31);
      x2 ^= x0;
      x1 = x0 << 3;
      x4 ^= x1;
      x0 = (x0 >>> 13) | (x0 << 19);
      x2 ^= x3;
      x4 ^= x3;
      x3 = (x3 >>> 3) | (x3 << 29);
      x3 ^= x4; x4 ^= x0;
      x1 = x4; x4 &= x3;
      x4 ^= x2; x2 |= x3;
      x2 ^= x1; x1 &= x4;
      x3 ^= x4; x1 &= x0;
      x1 ^= x3; x3 &= x2;
      x3 |= x0; x4 =~ x4;
      x3 ^= x4; x0 ^= x4;
      x0 &= x2; x4 ^= x1;
      x4 ^= x0;
      x2 ^= key.k8;
      x1 ^= key.k9;
      x3 ^= key.k10;
      x4 ^= key.k11;
      x3 = (x3 >>> 22) | (x3 << 10);
      x2 = (x2 >>> 5) | (x2 << 27);
      x0 = x1;
      x3 ^= x4;
      x0 <<= 7;
      x2 ^= x4;
      x3 ^= x0;
      x2 ^= x1;
      x4 = (x4 >>> 7) | (x4 << 25);
      x1 = (x1 >>> 1) | (x1 << 31);
      x1 ^= x2;
      x0 = x2 << 3;
      x4 ^= x0;
      x2 = (x2 >>> 13) | (x2 << 19);
      x1 ^= x3;
      x4 ^= x3;
      x3 = (x3 >>> 3) | (x3 << 29);
      x0 = x1; x1 ^= x4;
      x4 &= x1; x0 ^= x3;
      x4 ^= x2; x2 |= x1;
      x3 ^= x4; x2 ^= x0;
      x2 |= x3; x1 ^= x4;
      x2 ^= x1; x1 |= x4;
      x1 ^= x2; x0 =~ x0;
      x0 ^= x1; x1 |= x2;
      x1 ^= x2;
      x1 |= x0;
      x4 ^= x1;
      x0 ^= key.k4;
      x2 ^= key.k5;
      x4 ^= key.k6;
      x3 ^= key.k7;
      x4 = (x4 >>> 22) | (x4 << 10);
      x0 = (x0 >>> 5) | (x0 << 27);
      x1 = x2;
      x4 ^= x3;
      x1 <<= 7;
      x0 ^= x3;
      x4 ^= x1;
      x0 ^= x2;
      x3 = (x3 >>> 7) | (x3 << 25);
      x2 = (x2 >>> 1) | (x2 << 31);
      x2 ^= x0;
      x1 = x0 << 3;
      x3 ^= x1;
      x0 = (x0 >>> 13) | (x0 << 19);
      x2 ^= x4;
      x3 ^= x4;
      x4 = (x4 >>> 3) | (x4 << 29);
      x4 =~ x4; x1 = x2;
      x2 |= x0; x1 =~ x1;
      x2 ^= x4; x4 |= x1;
      x2 ^= x3; x0 ^= x1;
      x4 ^= x0; x0 &= x3;
      x1 ^= x0; x0 |= x2;
      x0 ^= x4; x3 ^= x1;
      x4 ^= x2; x3 ^= x0;
      x3 ^= x2;
      x4 &= x3;
      x1 ^= x4;
      x0 ^= key.k0;
      x1 ^= key.k1;
      x2 ^= key.k2;
      x3 ^= key.k3;

      out[o++] = (byte) (x3 >>> 24);
      out[o++] = (byte) (x3 >>> 16);
      out[o++] = (byte) (x3 >>>  8);
      out[o++] = (byte)  x3;
      out[o++] = (byte) (x2 >>> 24);
      out[o++] = (byte) (x2 >>> 16);
      out[o++] = (byte) (x2 >>>  8);
      out[o++] = (byte)  x2;
      out[o++] = (byte) (x1 >>> 24);
      out[o++] = (byte) (x1 >>> 16);
      out[o++] = (byte) (x1 >>>  8);
      out[o++] = (byte)  x1;
      out[o++] = (byte) (x0 >>> 24);
      out[o++] = (byte) (x0 >>> 16);
      out[o++] = (byte) (x0 >>>  8);
      out[o  ] = (byte)  x0;
   }

   public boolean selfTest() {
      if (valid == null) {
         boolean result = super.selfTest(); // do symmetry tests
         if (result) {
            System.out.println("passed symmetry test");
            result = testKat(KAT_KEY, KAT_CT);
         }
         valid = new Boolean(result);
      }
      return valid.booleanValue();
   }

   // Own methods.
   // -------------------------------------------------------------------------

   // Bit-flip madness methods
   //
   // The following S-Box functions were developed by Dag Arne Osvik, and are
   // described in his paper, "Speeding up Serpent". They are optimized to
   // perform on the Pentium chips, but work well here too.
   //
   // The methods below are Copyright (C) 2000 Dag Arne Osvik.

   /** S-Box 0. */
   private final void
   sbox0(int r0, int r1, int r2, int r3) {
      int r4 = r1 ^ r2;
      r3 ^= r0;
      r1 = r1 & r3 ^ r0;
      r0 = (r0 | r3) ^ r4;
      r4 ^= r3;
      r3 ^= r2;
      r2 = (r2 | r1) ^ r4;
      r4 = ~r4 | r1;
      r1 ^= r3 ^ r4;
      r3 |= r0;
      x0 = r1 ^ r3;
      x1 = r4 ^ r3;
      x2 = r2;
      x3 = r0;
   }

   /** S-Box 1. */
   private final void
   sbox1(int r0, int r1, int r2, int r3) {
      r0 = ~r0;
      int r4 = r0;
      r2 = ~r2;
      r0 &= r1;
      r2 ^= r0;
      r0 |= r3;
      r3 ^= r2;
      r1 ^= r0;
      r0 ^= r4;
      r4 |= r1;
      r1 ^= r3;
      r2 = (r2 | r0) & r4;
      r0 ^= r1;
      x0 = r2;
      x1 = r0 & r2 ^ r4;
      x2 = r3;
      x3 = r1 & r2 ^ r0;
   }

   /** S-Box 2. */
   private final void
   sbox2(int r0, int r1, int r2, int r3) {
      int r4 = r0;
      r0 = r0 & r2 ^ r3;
      r2 = r2 ^ r1 ^ r0;
      r3 = (r3 | r4) ^ r1;
      r4 ^= r2;
      r1 = r3;
      r3 = (r3 | r4) ^ r0;
      r0 &= r1;
      r4 ^= r0;
      x0 = r2;
      x1 = r3;
      x2 = r1 ^ r3 ^ r4;
      x3 = ~r4;
   }

   /** S-Box 3. */
   private final void
   sbox3(int r0, int r1, int r2, int r3) {
      int r4 = r0;
      r0 |= r3;
      r3 ^= r1;
      r1 &= r4;
      r4 = r4 ^ r2 | r1;
      r2 ^= r3;
      r3 = r3 & r0 ^ r4;
      r0 ^= r1;
      r4 = r4 & r0 ^ r2;
      r1 = (r1 ^ r3 | r0) ^ r2;
      r0 ^= r3;
      x0 = (r1 | r3) ^ r0;
      x1 = r1;
      x2 = r3;
      x3 = r4;
   }

   /** S-Box 4. */
   private final void
   sbox4(int r0, int r1, int r2, int r3) {
      r1 ^= r3;
      int r4 = r1;
      r3 = ~r3;
      r2 ^= r3;
      r3 ^= r0;
      r1 = r1 & r3 ^ r2;
      r4 ^= r3;
      r0 ^= r4;
      r2 = r2 & r4 ^ r0;
      r0 &= r1;
      r3 ^= r0;
      r4 = (r4 | r1) ^ r0;
      x0 = r1;
      x1 = r4 ^ (r2 & r3);
      x2 = ~((r0 | r3) ^ r2);
      x3 = r3;
   }

   /** S-Box 5. */
   private final void
   sbox5(int r0, int r1, int r2, int r3) {
      r0 ^= r1;
      r1 ^= r3;
      int r4 = r1;
      r3 = ~r3;
      r1 &= r0;
      r2 ^= r3;
      r1 ^= r2;
      r2 |= r4;
      r4 ^= r3;
      r3 = r3 & r1 ^ r0;
      r4 = r4 ^ r1 ^ r2;
      x0 = r1;
      x1 = r3;
      x2 = r0 & r3 ^ r4;
      x3 = ~(r2 ^ r0) ^ (r4 | r3);
   }

   /** S-Box 6. */
   private final void
   sbox6(int r0, int r1, int r2, int r3) {
      int r4 = r3;
      r2 = ~r2;
      r3 = r3 & r0 ^ r2;
      r0 ^= r4;
      r2 = (r2 | r4) ^ r0;
      r1 ^= r3;
      r0 |= r1;
      r2 ^= r1;
      r4 ^= r0;
      r0 = (r0 | r3) ^ r2;
      r4 = r4 ^ r3 ^ r0;
      x0 = r0;
      x1 = r1;
      x2 = r4;
      x3 = r2 & r4 ^ ~r3;
   }

   /** S-Box 7. */
   private final void
   sbox7(int r0, int r1, int r2, int r3) {
      int r4 = r1;
      r1 = (r1 | r2) ^ r3;
      r4 ^= r2;
      r2 ^= r1;
      r3 = (r3 | r4) & r0;
      r4 ^= r2;
      r3 ^= r1;
      r1 = (r1 | r4) ^ r0;
      r0 = (r0 | r4) ^ r2;
      r1 ^= r4;
      r2 ^= r1;
      x0 = r4 ^ (~r2 | r0);
      x1 = r3;
      x2 = r1 & r0 ^ r4;
      x3 = r0;
   }
}

